# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import annotations

import json
import re

from extra_platforms import ALL_PLATFORMS

from ..base import PackageManager
from ..capabilities import search_capabilities, version_not_implemented

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Iterator

    from ..base import Package


class Composer(PackageManager):
    name = "PHP's Composer"

    homepage_url = "https://getcomposer.org"

    platforms = ALL_PLATFORMS

    requirement = "1.4.0"

    _SEARCH_REGEXP = re.compile(
        r"""
        ^(?P<package_id>\S+\/\S+)
        (?:\s+(?P<description>\S.*))?
        """,
        re.MULTILINE | re.VERBOSE,
    )

    pre_args = ("global", "--no-ansi")

    version_regexes = (r"Composer\s+version\s+(?P<version>\S+)",)
    """
    .. code-block:: shell-session

        $ composer --version
        Composer version 2.1.8 2021-09-15 13:55:14
    """

    @property
    def installed(self) -> Iterator[Package]:
        """Fetch installed packages.

        .. code-block:: shell-session

            $ composer global show --format=json | jq
            {
              "installed": [
                {
                  "name": "carbondate/carbon",
                  "version": "1.33.0",
                  "description": "A simple API extension for DateTime."
                },
                {
                  "name": "guzzlehttp/guzzle",
                  "version": "6.3.3",
                  "description": "Guzzle is a PHP HTTP client library"
                },
                {
                  "name": "guzzlehttp/promises",
                  "version": "v1.3.1",
                  "description": "Guzzle promises library"
                },
                {
                  "name": "guzzlehttp/psr7",
                  "version": "1.4.2",
                  "description": "PSR-7 message (...) methods"
                },
            (...)
        """
        output = self.run_cli("show", "--format=json")
        if output:
            package_list = json.loads(output)
            for package in package_list["installed"]:
                package_id = package["name"]
                yield self.package(id=package_id, installed_version=package["version"])

    @property
    def outdated(self) -> Iterator[Package]:
        """Fetch outdated packages.

        .. code-block:: shell-session

            $ composer global outdated --format=json
            {
                "installed": [
                    {
                        "name": "illuminate/contracts",
                        "version": "v5.7.2",
                        "latest": "v5.7.3",
                        "latest-status": "semver-safe-update",
                        "description": "The Illuminate Contracts package."
                    },
                    {
                        "name": "illuminate/support",
                        "version": "v5.7.2",
                        "latest": "v5.7.3",
                        "latest-status": "semver-safe-update",
                        "description": "The Illuminate Support package."
                    }
                ]
            }
        """
        output = self.run_cli("outdated", "--format=json")

        if output:
            package_list = json.loads(output)
            for package in package_list["installed"]:
                package_id = package["name"]
                yield self.package(
                    id=package_id,
                    installed_version=package["version"],
                    latest_version=package["latest"],
                )

    @search_capabilities(exact_support=False)
    def search(self, query: str, extended: bool, exact: bool) -> Iterator[Package]:
        """Fetch matching packages.

        .. caution::
            Search does not supports exact matching.

        .. code-block:: shell-session

            $ composer global search symfony
            symfony/symfony The Symfony PHP framework
            symfony/yaml Symfony Yaml Component
            symfony/var-dumper Symfony (...) dumping PHP variables
            symfony/translation Symfony Translation Component
            symfony/routing Symfony Routing Component
            symfony/process Symfony Process Component
            symfony/polyfill-php70 Symfony (...) features to lower PHP versions
            symfony/polyfill-mbstring Symfony (...) Mbstring extension
            symfony/polyfill-ctype Symfony polyfill for ctype functions
            symfony/http-kernel Symfony HttpKernel Component
            symfony/http-foundation Symfony HttpFoundation Component
            symfony/finder Symfony Finder Component
            symfony/event-dispatcher Symfony EventDispatcher Component
            symfony/debug Symfony Debug Component
            symfony/css-selector Symfony CssSelector Component

        .. code-block:: shell-session

            $ composer global search --only-name python
            hiqdev/hidev-python
            aanro/pythondocx
            laravel-admin-ext/python-editor
            pythonphp/pythonphp
            blyxxyz/python-server
            nim-development/python-domotics
            rakshitbharat/pythoninphp
            tequilarapido/python-bridge

        .. code-block:: shell-session

            $ search global --only-name pythonphp/pythonphp
            pythonphp/pythonphp
        """
        search_args = []
        if not extended:
            search_args.append("--only-name")

        output = self.run_cli("search", search_args, query)

        for match in self._SEARCH_REGEXP.finditer(output):
            package_id = match.group("package_id")
            description = match.group("description")
            yield self.package(id=package_id, description=description)

    @version_not_implemented
    def install(self, package_id: str, version: str | None = None) -> str:
        """Install one package.

        .. code-block:: shell-session

            $ composer global install illuminate/contracts
        """
        return self.run_cli("install", package_id)

    def upgrade_all_cli(self) -> tuple[str, ...]:
        """Generates the CLI to upgrade all packages (default) or only the one provided
        as parameter.

        .. code-block:: shell-session

            $ composer global update
        """
        return self.build_cli("update")

    @version_not_implemented
    def upgrade_one_cli(
        self,
        package_id: str,
        version: str | None = None,
    ) -> tuple[str, ...]:
        """Generates the CLI to upgrade all packages (default) or only the one provided
        as parameter.

        .. code-block:: shell-session

            $ composer global update illuminate/contracts
        """
        return self.build_cli("update", package_id)

    def cleanup(self) -> None:
        """Removes things we don't need anymore.

        See: https://getcomposer.org/doc/03-cli.md#clear-cache-clearcache-cc

        .. code-block:: shell-session

            $ composer global clear-cache
        """
        self.run_cli("clear-cache")



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import annotations

import json
import logging
import re
from operator import methodcaller

from extra_platforms import LINUX_LIKE, MACOS

from ..base import PackageManager
from ..capabilities import version_not_implemented
from ..version import parse_version

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Iterator

    from ..base import Package


class Homebrew(PackageManager):
    """Virtual package manager shared by brew and cask CLI defined below.

    Homebrew is the umbrella project providing both brew and brew cask commands.
    """

    platforms = LINUX_LIKE, MACOS
    """Homebrew core is now compatible with `Linux and Windows Subsystem for Linux (WSL)
    2 <https://docs.brew.sh/Homebrew-on-Linux>`_."""

    requirement = "2.7.0"
    """Vanilla ``brew`` and ``cask`` CLIs now shares the same version.

    `2.7.0 <https://github.com/Homebrew/brew/releases/tag/2.7.0>`_ is the first release
    to enforce the use of ``--cask`` option.
    """

    # Declare this manager as virtual, i.e. not tied to a real CLI.
    virtual = True

    extra_env = {  # noqa: RUF012
        # Disable analytics.
        "HOMEBREW_NO_ANALYTICS": "1",
        # Disable configuration hints to reduce verbosity.
        "HOMEBREW_NO_ENV_HINTS": "1",
        # Do not let brew mix the update operation with others. Mpm has a separate
        # "sync" command for that. This silo-ed behavior has been requested by user
        # since the beginning of mpm:
        # https://github.com/kdeldycke/meta-package-manager/issues/36
        "HOMEBREW_NO_AUTO_UPDATE": "1",
        # See: https://docs.brew.sh/FAQ#why-cant-i-open-a-mac-app-from-an-unidentified-developer
        # "HOMEBREW_CASK_OPTS": "--no-quarantine",
    }

    _INSTALLED_REGEXP = re.compile(
        r"""
        (?P<package_id>\S+)     # Any non-empty characters.
        (?P<removed> \(!\))?    # Package removed flag.
        \                       # A space.
        (?P<versions>.+)        # Versions.
        """,
        re.VERBOSE,
    )
    _SEARCH_DESC_REGEXP = re.compile(
        r"""
        (?:==>\s\S+\s)?           # Ignore section starting with '==>'.
        (?P<package_id>\S+)       # Any non-empty characters.
        :                         # Semi-colon.
        (                         # Optional group start (ignored below with _).
            \s+                   # Blank characters.
            \(                    # Opening parenthesis.
            (?P<package_name>.+)  # Any string.
            \)                    # Closing parenthesis.
        )?                        # Optional group end.
        \s+                       # Blank characters.
        (?P<description>.+)       # Any string.
        """,
        re.VERBOSE,
    )
    _SEARCH_REGEXP = re.compile(
        r"""
        (?:==>\s\S+\s)?           # Ignore section starting with '==>'.
        (?P<package_id>[^\s✔]+)   # Anything not a whitespace or ✔.
        """,
        re.VERBOSE,
    )

    version_regexes = (r"Homebrew\s+(?P<version>\S+)",)
    """
    .. code-block:: shell-session

        $ brew --version
        Homebrew 1.8.6-124-g6cd4c31
        Homebrew/homebrew-core (git revision 533d; last commit 2018-12-28)
        Homebrew/homebrew-cask (git revision 5095b; last commit 2018-12-28)
    """

    @property
    def installed(self) -> Iterator[Package]:
        """Fetch installed packages.

        .. code-block:: shell-session

            $ brew list --versions --formula
            ack 2.14
            apg 2.2.3
            audacity (!) 2.1.2
            apple-gcc42 4.2.1-5666.3
            atk 2.22.0
            bash 4.4.5
            bash-completion 1.3_1
            boost 1.63.0
            c-ares 1.12.0
            graphviz 2.40.1 2.40.20161221.0239
            quicklook-json latest

        .. code-block:: shell-session

            $ brew list --versions --cask
            aerial 1.2beta5
            android-file-transfer latest
            audacity (!) 2.1.2
            firefox 49.0.1
            flux 37.7
            gimp 2.8.18-x86_64
            java 1.8.0_112-b16
            tunnelblick 3.6.8_build_4625 3.6.9_build_4685
            virtualbox 5.1.8-111374 5.1.10-112026

        .. todo::

            Use the ``removed`` variable to detect removed packages (which are
            reported with a ``(!)`` flag). See:
            https://github.com/caskroom/homebrew-cask/blob/master/doc
            /reporting_bugs/uninstall_wrongly_reports_cask_as_not_installed.md
            and https://github.com/kdeldycke/meta-package-manager/issues/17 .
        """
        output = self.run_cli("list", "--quiet", "--versions")

        for package_id, _removed, versions in map(
            methodcaller("groups"),
            self._INSTALLED_REGEXP.finditer(output),
        ):
            # Keep highest version found.
            version = max(map(parse_version, versions.split()))
            yield self.package(id=package_id, installed_version=version)

    @property
    def outdated(self) -> Iterator[Package]:
        """Fetch outdated packages.

        .. code-block:: shell-session

            $ brew outdated --json=v2 --formula | jq
            {
              "formulae": [
                {
                  "name": "pygobject3",
                  "installed_versions": [
                    "3.36.1"
                  ],
                  "current_version": "3.38.0",
                  "pinned": false,
                  "pinned_version": null
                },
                {
                  "name": "rav1e",
                  "installed_versions": [
                    "0.3.3"
                  ],
                  "current_version": "0.3.4",
                  "pinned": false,
                  "pinned_version": null
                }
              ],
              "casks": []
            }

        .. code-block:: shell-session

            $ brew outdated --json=v2 --cask | jq
            {
              "formulae": [],
              "casks": [
                {
                  "name": "electrum",
                  "installed_versions": "4.0.2",
                  "current_version": "4.0.3"
                },
                {
                  "name": "qlcolorcode",
                  "installed_versions": "3.0.2",
                  "current_version": "3.1.1"
                }
              ]
            }

        .. code-block:: shell-session

            $ brew outdated --json=v2 --greedy --cask | jq
            {
              "formulae": [],
              "casks": [
                {
                  "name": "amethyst",
                  "installed_versions": "0.14.3",
                  "current_version": "0.15.3"
                },
                {
                  "name": "balenaetcher",
                  "installed_versions": "1.5.106",
                  "current_version": "1.5.108"
                },
                {
                  "name": "caldigit-thunderbolt-charging",
                  "installed_versions": "latest",
                  "current_version": "latest"
                },
                {
                  "name": "electrum",
                  "installed_versions": "4.0.2",
                  "current_version": "4.0.3"
                },
                {
                  "name": "lg-onscreen-control",
                  "installed_versions": "5.33,cV8xqv5TSZA.upgrading, 5.47,yi5XuIZw6hg",
                  "current_version": "5.48,uYXSwyUCNFBbSch9PFw"
                }
              ]
            }
        """
        # Build up the list of CLI options.
        options = ["--json=v2"]
        # Includes auto-update packages or not.
        if not self.ignore_auto_updates:
            options.append("--greedy")

        # List available updates. Do not use --quiet here: brew treats --quiet
        # and --json as mutually exclusive. See:
        # https://github.com/kdeldycke/meta-package-manager/issues/1703
        output = self.run_cli("outdated", options)

        if output:
            package_list = json.loads(output)
            for pkg_info in package_list["formulae"] + package_list["casks"]:
                # Interpret installed versions.
                versions = pkg_info["installed_versions"]
                if isinstance(versions, str):
                    versions = versions.split(", ")
                installed_version = max(map(parse_version, versions))

                latest_version = parse_version(pkg_info["current_version"])

                # Skip packages not offering upgradeable version.
                package_id = pkg_info["name"]
                if installed_version == latest_version:
                    logging.debug(
                        f"Ignore {package_id} upgrade "
                        f"from {installed_version} to {latest_version}.",
                    )
                    continue

                yield self.package(
                    id=package_id,
                    installed_version=installed_version,
                    latest_version=latest_version,
                )

    def search(self, query: str, extended: bool, exact: bool) -> Iterator[Package]:
        """Fetch matching packages.

        .. caution::
            Search does not supports extended mode.

        .. code-block:: shell-session

            $ brew search sed
            ==> Formulae
            gnu-sed ✔                    libxdg-basedir
            ==> Casks
            eclipse-dsl                       marsedit
            focused                           physicseditor
            google-adwords-editor             prefs-editor
            licensed                          subclassed-mnemosyne

        .. code-block:: shell-session

            $ brew search sed --formulae
            ==> Formulae
            gnu-sed ✔                    libxdg-basedir

        .. code-block:: shell-session

            $ brew search sed --cask
            ==> Casks
            eclipse-dsl                       marsedit
            focused                           physicseditor
            google-adwords-editor             prefs-editor
            licensed                          subclassed-mnemosyne

        .. code-block:: shell-session

            $ brew search python --formulae
            ==> Formulae
            app-engine-python   boost-python3   python ✔          python-yq
            boost-python        gst-python      python-markdown   python@3.8 ✔

        .. code-block:: shell-session

            $ brew search "/^ssed$/" --formulae
            ==> Formulae
            ssed

        .. code-block:: shell-session

            $ brew search "/^sed$/" --formulae
            Error: No formula or cask found for "/^sed$/".

        .. code-block:: shell-session

            $ brew search tetris --formulae --desc
            ==> Formulae
            bastet: Bastard Tetris
            netris: Networked variant of tetris
            vitetris: Terminal-based Tetris clone
            yetris: Customizable Tetris for the terminal

        .. code-block:: shell-session

            $ brew search tetris --cask --desc
            ==> Casks
            not-tetris: (Not Tetris) [no description]
            tetrio: (TETR.IO) Free-to-play Tetris clone

        More doc at: https://docs.brew.sh/Manpage#search--s-options-textregex-
        """
        # Keep track of package IDs already matched by the first extended search pass.
        matched_ids = set()

        # Additional search on description only.
        if extended:
            output = self.run_cli("search", "--quiet", query, "--desc")

            for package_id, _, package_name, description in (
                self._SEARCH_DESC_REGEXP.findall(output)
            ):
                matched_ids.add(package_id)
                pkg = self.package(id=package_id, name=package_name)
                if description != "[no description]":
                    pkg.description = description
                yield pkg

        # Use regexp if exact match is requested.
        if exact:
            query = f"/^{query}$/"

        output = self.run_cli("search", "--quiet", query)

        for package_id in self._SEARCH_REGEXP.findall(output):
            # Deduplicate search results.
            if package_id not in matched_ids:
                yield self.package(id=package_id)

    @version_not_implemented
    def install(self, package_id: str, version: str | None = None) -> str:
        """Install one package.

        .. code-block:: shell-session

            $ brew install jpeginfo --formula
            ==> Downloading https://ghcr.io/core/jpeginfo/manifests/1.6.1_1-1
            ############################################################## 100.0%
            ==> Downloading https://ghcr.io/core/jpeginfo/blobs/sha256:27bb35884368b83
            ==> Downloading from https://pkg.githubcontent.com/ghcr1/blobs/sha256:27bb3
            ############################################################## 100.0%
            ==> Pouring jpeginfo--1.6.1_1.big_sure.bottle.1.tar.gz
            🍺  /usr/local/Cellar/jpeginfo/1.6.1_1: 7 files, 77.6KB

        .. code-block:: shell-session

            $ brew install pngyu --cask
            ==> Downloading https://nukesaq.github.io/Pngyu/download/Pngyu_mac_101.zip
            ################################################################## 100.0%
            ==> Installing Cask pngyu
            ==> Moving App 'Pngyu.app' to '/Applications/Pngyu.app'
            🍺  pngyu was successfully installed!
        """
        return self.run_cli("install", "--quiet", package_id)

    def upgrade_all_cli(self) -> tuple[str, ...]:
        """Generates the CLI to upgrade all packages (default) or only the one provided
        as parameter.

        ``brew`` and ``cask`` share the same command.

        .. code-block:: shell-session

            $ brew upgrade --formula
            ==> Upgrading 2 outdated packages:
            node 13.11.0 -> 13.12.0
            sdl2 2.0.12 -> 2.0.12_1
            ==> Upgrading node 13.11.0 -> 13.12.0
            ==> Downloading https://homebrew.bintray.com/bottles/node-13.tar.gz
            ==> Downloading from https://akamai.bintray.com/fc/fc0bfb42fe23e960
            ############################################################ 100.0%
            ==> Pouring node-13.12.0.catalina.bottle.tar.gz
            ==> Caveats
            Bash completion has been installed to:
              /usr/local/etc/bash_completion.d
            ==> Summary
            🍺  /usr/local/Cellar/node/13.12.0: 4,660 files, 60.3MB
            Removing: /usr/local/Cellar/node/13.11.0... (4,686 files, 60.4MB)
            ==> Upgrading sdl2 2.0.12 -> 2.0.12_1
            ==> Downloading https://homebrew.bintray.com/bottles/sdl2-2.tar.gz
            ==> Downloading from https://akamai.bintray.com/4d/4dcd635465d16372
            ############################################################ 100.0%
            ==> Pouring sdl2-2.0.12_1.catalina.bottle.tar.gz
            🍺  /usr/local/Cellar/sdl2/2.0.12_1: 89 files, 4.7MB
            Removing: /usr/local/Cellar/sdl2/2.0.12... (89 files, 4.7MB)
            ==> Checking for dependents of upgraded formulae...
            ==> No dependents found!
            ==> Caveats
            ==> node
            Bash completion has been installed to:
              /usr/local/etc/bash_completion.d

        .. code-block:: shell-session

            $ brew upgrade --cask
            ==> Casks with `auto_updates` or `version :latest` will not be upgraded
            ==> Upgrading 1 outdated packages:
            aerial 2.0.7 -> 2.0.8
            ==> Upgrading aerial
            ==> Downloading https://github.com/Aerial/download/v2.0.8/Aerial.saver.zip
            ==> Downloading from https://65be.s3.amazonaws.com/44998092/29eb1e0
            ==> Verifying SHA-256 checksum for Cask 'aerial'.
            ==> Backing Screen Saver up to '/usr/local/Caskroom/Aerial.saver'.
            ==> Removing Screen Saver '/Users/kde/Library/Screen Savers/Aerial.saver'.
            ==> Moving Screen Saver to '/Users/kde/Library/Screen Savers/Aerial.saver'.
            ==> Purging files for version 2.0.7 of Cask aerial
            🍺  aerial was successfully upgraded!
        """
        return self.build_cli("upgrade", "--quiet")

    @version_not_implemented
    def upgrade_one_cli(
        self,
        package_id: str,
        version: str | None = None,
    ) -> tuple[str, ...]:
        """Generates the CLI to upgrade all packages (default) or only the one provided
        as parameter.

        ``brew`` and ``cask`` share the same command.

        .. code-block:: shell-session

            $ brew upgrade dupeguru --cask
            ==> Upgrading 1 outdated package:
            dupeguru 4.2.0 -> 4.2.1
            ==> Upgrading dupeguru
            ==> Downloading https://github.com/(...)/4.2.1/dupeguru_macOS_Qt_4.2.1.zip
            ==> Downloading from https://githubusercontent.com/production-release-asset
            ##################################################################### 100.0%
            ==> Backing App 'dupeguru.app' up to '/opt/homebrew/.../4.2.0/dupeguru.app'
            ==> Removing App '/Applications/dupeguru.app'
            ==> Moving App 'dupeguru.app' to '/Applications/dupeguru.app'
            ==> Purging files for version 4.2.0 of Cask dupeguru
            🍺  dupeguru was successfully upgraded!
        """
        return self.build_cli("upgrade", "--quiet", package_id)

    def remove(self, package_id: str) -> str:
        """Removes a package.

        .. code-block:: shell-session

            $ brew uninstall bat
            Uninstalling /usr/local/Cellar/bat/0.21.0... (14 files, 5MB)
        """
        return self.run_cli("uninstall", "--quiet", package_id)

    def sync(self) -> None:
        """Sync package metadata.

        .. code-block:: shell-session

            $ brew update --quiet
            Already up-to-date.
        """
        self.run_cli("update", "--quiet", auto_post_args=False)

    def cleanup(self) -> None:
        """Removes things we don't need anymore.

        Scrub the cache, including latest version's downloads. Also remove unused
        dependencies.

        Downloads for all installed formulae and casks will not be deleted.

        .. code-block:: shell-session

            $ brew cleanup -s --prune=all
            Removing: ~/Library/Caches/Homebrew/node--1.bottle.tar.gz... (9MB)
            Warning: Skipping sdl2: most recent version 2.0.12_1 not installed
            Removing: ~/Library/Caches/Homebrew/Cask/aerial--1.8.1.zip... (5MB)
            Removing: ~/Library/Caches/Homebrew/Cask/prey--1.9.pkg... (19.9MB)
            Removing: ~/Library/Logs/Homebrew/readline... (64B)
            Removing: ~/Library/Logs/Homebrew/libfido2... (64B)
            Removing: ~/Library/Logs/Homebrew/libcbor... (64B)

        More doc at: https://docs.brew.sh/Manpage#cleanup-options-formulacask

        .. code-block:: shell-session

            $ brew autoremove
            ==> Uninstalling 17 unneeded formulae:
            gtkmm3
            highlight
            lua@5.1
            nasm
            nghttp2
            texi2html
            Uninstalling /usr/local/Cellar/nghttp2/1.41.0_1... (26 files, 2.7MB)
            Uninstalling /usr/local/Cellar/highlight/3.59... (558 files, 3.5MB)

            Warning: The following highlight configuration files have not been removed!
            If desired, remove them manually with `rm -rf`:
              /usr/local/etc/highlight
              /usr/local/etc/highlight/filetypes.conf
              /usr/local/etc/highlight/filetypes.conf.default
            Uninstalling /usr/local/Cellar/gtkmm3/3.24.2_1... (1,903 files, 173.7MB)
            Uninstalling /usr/local/Cellar/texi2html/5.0... (279 files, 6.2MB)
            Uninstalling /usr/local/Cellar/lua@5.1/5.1.5_8... (22 files, 245.6KB)
            Uninstalling /usr/local/Cellar/nasm/2.15.05... (29 files, 2.9MB)
        """
        self.run_cli("autoremove", "--quiet", auto_post_args=False)
        self.run_cli("cleanup", "--quiet", "-s", "--prune=all", auto_post_args=False)


class Brew(Homebrew):
    name = "Homebrew Formulae"

    homepage_url = "https://brew.sh"

    cli_names = ("brew",)

    post_args = ("--formula",)


class Cask(Homebrew):
    name = "Homebrew Cask"

    homepage_url = "https://github.com/Homebrew/homebrew-cask"

    platforms = MACOS  # type: ignore[assignment]
    """Casks are only available on macOS, not Linux or WSL."""

    cli_names = ("brew",)

    post_args = ("--cask",)



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import annotations

import json
import re
from functools import cached_property

from extra_platforms import ALL_PLATFORMS

from ..base import PackageManager
from ..capabilities import search_capabilities, version_not_implemented

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Iterator

    from ..base import Package


class Yarn(PackageManager):
    """See command equivalences at: https://github.com/antfu-collective/ni?tab=readme-ov-file#ni."""

    name = "Node's yarn"

    homepage_url = "https://yarnpkg.com"

    platforms = ALL_PLATFORMS

    requirement = "1.20.0"

    pre_args = ("--silent",)

    _INSTALLED_REGEXP = re.compile(
        r"^.+\"data\":\"\\\"(?P<package_id>\S+)"
        r"@(?P<version>\S+)\\\" has binaries:\"\}$",
        re.MULTILINE,
    )

    """
    .. code-block:: shell-session

        $ yarn --version
        1.22.11
    """

    @property
    def installed(self) -> Iterator[Package]:
        """Fetch installed packages.

        .. code-block:: shell-session

            $ yarn global --json list --depth 0
            {"type":"activityStart","data":{"id":0}}
            {"type":"activityTick","data":{"id":0,"name":"awesome-lint@^0.18.0"}}
            {"type":"activityTick","data":{"id":0,"name":"arrify@^2.0.1"}}
            {"type":"activityTick","data":{"id":0,"name":"case@^1.6.3"}}
            {"type":"activityTick","data":{"id":0,"name":"emoji-regex@^9.2.0"}}
            (...)
            {"type":"activityEnd","data":{"id":0}}
            {"type":"progressStart","data":{"id":0,"total":327}}
            {"type":"progressTick","data":{"id":0,"current":1}}
            {"type":"progressTick","data":{"id":0,"current":2}}
            {"type":"progressTick","data":{"id":0,"current":3}}
            {"type":"progressTick","data":{"id":0,"current":4}}
            {"type":"progressTick","data":{"id":0,"current":5}}
            (...)
            {"type":"progressFinish","data":{"id":0}}
            {"type":"info","data":"\"awesome-lint@0.18.0\" has binaries:"}
            {"type":"list","data":{"type":"bins-awesome-lint","items":["awesome-lint"]}}

        .. code-block:: shell-session

            $ yarn global list --depth 0
            yarn global v1.22.19
            info "awesome-lint@0.18.0" has binaries:
               - awesome-lint
            ✨  Done in 0.13s.
        """
        output = self.run_cli("global", "--json", "list", "--depth", "0")

        for package_id, version in self._INSTALLED_REGEXP.findall(output):
            yield self.package(id=package_id, installed_version=version)

    @cached_property
    def global_dir(self) -> str:
        """Locate the global directory.

        .. code-block:: shell-session

            $ yarn global dir
            ~/.config/yarn/global
        """
        return self.run_cli("global", "dir", force_exec=True).rstrip()

    @property
    def outdated(self) -> Iterator[Package]:
        """Fetch outdated packages.

        .. code-block:: shell-session

            $ yarn --json outdated --cwd ~/.config/yarn/global | jq
            {"type":"warning","data":"package.json: No license field"}
            {
              "type": "info",
              "data":
                "Color legend : \n"
                " \"<red>\"    : Major Update backward-incompatible updates \n"
                " \"<yellow>\" : Minor Update backward-compatible features \n"
                " \"<green>\"  : Patch Update backward-compatible bug fixes"
            }
            {
              "type": "table",
              "data": {
                "head": [
                  "Package",
                  "Current",
                  "Wanted",
                  "Latest",
                  "Package Type",
                  "URL"
                ],
                "body": [
                  [
                    "markdown",
                    "0.4.0",
                    "0.4.0",
                    "0.5.0",
                    "dependencies",
                    "git://github.com/evilstreak/markdown-js.git"
                  ]
                ]
              }
            }

        .. code-block:: shell-session

            $ yarn outdated --cwd ~/.config/yarn/global
            yarn outdated v1.22.19
            warning package.json: No license field
            info Color legend :
            "<red>"    : Major Update backward-incompatible updates
            "<yellow>" : Minor Update backward-compatible features
            "<green>"  : Patch Update backward-compatible bug fixes
            Package  Current Wanted Latest Package Type URL
            markdown 0.4.0   0.4.0  0.5.0  dependencies git://github.com/.../md-js.git
            ✨  Done in 0.95s.
        """
        output = self.run_cli("--json", "outdated", "--cwd", self.global_dir)
        if output:
            for line in output.splitlines():
                if not line:
                    continue
                obj = json.loads(line)
                if obj["type"] == "table":
                    for package in obj["data"]["body"]:
                        if package[2] == "linked":
                            continue
                        yield self.package(
                            id=package[0],
                            installed_version=package[1],
                            latest_version=package[3],
                        )

    @search_capabilities(extended_support=False, exact_support=False)
    def search(self, query: str, extended: bool, exact: bool) -> Iterator[Package]:
        """Fetch matching packages.

        .. warning::
            Yarn maintainers have `decided to not implement a dedicated search command
            <https://github.com/yarnpkg/yarn/issues/778#issuecomment-253146299>`_.

            Search is simulated by a direct call to ``yarn info``, and as a result only
            works for exact match.

        .. code-block:: shell-session

            $ yarn --json info python | jq
            {
              "type": "inspect",
              "data": {
                "name": "python",
                "description": "Interact with python child process",
                "dist-tags": {
                  "latest": "0.0.4"
                },
                "versions": [
                  "0.0.0",
                  "0.0.1",
                  "0.0.2",
                  "0.0.3",
                  "0.0.4"
                ],
                "maintainers": [
                  {
                    "name": "drderidder",
                    "email": "drderidder@gmail.com"
                  }
                ],
                "time": {
                  "modified": "2017-09-16T05:26:13.151Z",
                  "created": "2011-07-11T01:59:04.362Z",
                  "0.0.0": "2011-07-11T01:59:05.137Z",
                  "0.0.1": "2011-07-17T05:23:33.166Z",
                  "0.0.2": "2011-07-20T03:42:50.379Z",
                  "0.0.3": "2014-06-08T00:39:08.562Z",
                  "0.0.4": "2015-01-25T02:48:07.820Z"
                },
                "author": {
                  "name": "Darren DeRidder"
                },
                "repository": {
                  "type": "git",
                  "url": "git://github.com/73rhodes/node-python.git"
                },
                "homepage": "https://github.com/73rhodes/node-python",
                "bugs": {
                  "url": "https://github.com/73rhodes/node-python/issues"
                },
                "readmeFilename": "README.md",
                "users": {
                  "dewang-mistry": true,
                  "goliatone": true,
                  "sapanbhuta": true,
                  "aditcmarix": true,
                  "imlucas": true,
                  "heyderpd": true,
                  "ukuli": true,
                  "chbardel": true,
                  "asaupup": true,
                  "nuwaio": true
                },
                "version": "0.0.4",
                "main": "./lib/python.js",
                "engines": {
                  "node": ">= 0.4.1"
                },
                "gitHead": "69754aaa57658193916a1bf5fc391198098f74f6",
                "scripts": {},
                "dist": {
                  "shasum": "3094e898ef17a33aa9c3e973b3848a38e47d1818",
                  "tarball": "https://registry.npmjs.org/python/-/python-1.tgz"
                },
                "directories": {}
              }
            }
        """
        output = self.run_cli("--json", "info", query)

        if output:
            result = json.loads(output)

            if result["type"] == "inspect":
                package = result["data"]
                yield self.package(
                    id=package["name"],
                    description=package["description"],
                    latest_version=package["version"],
                )

    @version_not_implemented
    def install(self, package_id: str, version: str | None = None) -> str:
        """Install one package.

        .. code-block:: shell-session

            $ yarn global add awesome-lint
            yarn global v1.22.19
            [1/4] 🔍  Resolving packages...
            [2/4] 🚚  Fetching packages...
            [3/4] 🔗  Linking dependencies...
            [4/4] 🔨  Building fresh packages...

            success Installed "awesome-lint@0.18.0" with binaries:
                - awesome-lint
            ✨  Done in 16.15s.
        """
        return self.run_cli("global", "add", package_id)

    def upgrade_all_cli(self) -> tuple[str, ...]:
        """Generates the CLI to upgrade all packages (default) or only the one provided
        as parameter.

        .. code-block:: shell-session

            $ yarn global upgrade --latest
            yarn global v1.22.19
            [1/4] 🔍  Resolving packages...
            [2/4] 🚚  Fetching packages...
            [3/4] 🔗  Linking dependencies...
            [4/4] 🔨  Rebuilding all packages...
            success Saved lockfile.
            success Saved 271 new dependencies.
            info Direct dependencies
            ├─ awesome-lint@0.18.0
            └─ markdown@0.5.0
            info All dependencies
            ├─ @babel/code-frame@7.18.6
            ├─ @babel/helper-validator-identifier@7.18.6
            ├─ @nodelib/fs.scandir@2.1.5
            ├─ array-to-sentence@1.1.0
            ├─ array-union@2.1.0
            ├─ awesome-lint@0.18.0
            ├─ fs.realpath@1.0.0
            (...)
            └─ zwitch@1.0.5
            ✨  Done in 19.89s.
        """
        return self.build_cli("global", "upgrade", "--latest")

    @version_not_implemented
    def upgrade_one_cli(
        self,
        package_id: str,
        version: str | None = None,
    ) -> tuple[str, ...]:
        """Generates the CLI to upgrade all packages (default) or only the one provided
        as parameter.

        .. code-block:: shell-session

            $ yarn global upgrade markdown --latest
            yarn global v1.22.19
            [1/4] 🔍  Resolving packages...
            [2/4] 🚚  Fetching packages...
            [3/4] 🔗  Linking dependencies...
            [4/4] 🔨  Rebuilding all packages...
            success Saved lockfile.
            success Saved 2 new dependencies.
            info Direct dependencies
            └─ markdown@0.5.0
            info All dependencies
            ├─ markdown@0.5.0
            └─ nopt@2.1.2
            ✨  Done in 1.77s.
        """
        return self.build_cli("global", "upgrade", package_id, "--latest")

    def remove(self, package_id: str) -> str:
        """Remove one package.

        .. code-block:: shell-session

            $ yarn global remove awesome-lint
            yarn global v1.22.19
            [1/2] 🗑  Removing module awesome-lint...
            [2/2] 🔨  Regenerating lockfile and installing missing dependencies...
            success Uninstalled packages.
            ✨  Done in 0.21s.
        """
        return self.run_cli("global", "remove", package_id)

    def cleanup(self) -> None:
        """Removes things we don't need anymore.

        See: https://yarnpkg.com/cli/cache/clean

        .. code-block:: shell-session

            $ yarn cache clean --all
            yarn cache v1.22.19
            success Cleared cache.
            ✨  Done in 0.35s.
        """
        self.run_cli("cache", "clean", "--all")



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import annotations

import re

from extra_platforms import ALL_PLATFORMS

from ..base import PackageManager
from ..capabilities import search_capabilities, version_not_implemented

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Iterator

    from ..base import Package


class Cargo(PackageManager):
    name = "Rust's cargo"

    homepage_url = "https://doc.rust-lang.org/cargo/"

    platforms = ALL_PLATFORMS

    requirement = "1.0.0"

    pre_args = (
        "--color",
        "never",  # Suppress colored output.
        "--quiet",  # Do not print cargo log messages.
    )

    _INSTALLED_REGEXP = re.compile(
        r"^(?P<package_id>\S+)\s+v(?P<package_version>\S+):$",
    )
    _SEARCH_REGEXP = re.compile(
        r"^(?P<package_id>\S+)\s+=\s+\"(?P<version>\S+)\"\s+#\s+(?P<description>.+)$",
        re.MULTILINE,
    )

    version_regexes = (r"cargo\s+(?P<version>\S+)",)
    """
    .. code-block:: shell-session

        $ cargo --version
        cargo 1.59.0
    """

    @property
    def installed(self) -> Iterator[Package]:
        """Fetch installed packages.

        .. code-block:: shell-session

            $ cargo --color never --quiet install --list
            bore-cli v0.4.0:
                bore
            ripgrep v13.0.0:
                rg
        """
        output = self.run_cli("install", "--list")

        for package in output.splitlines():
            match = self._INSTALLED_REGEXP.match(package)
            if match:
                package_id, package_version = match.groups()
                yield self.package(id=package_id, installed_version=package_version)

    @search_capabilities(extended_support=False, exact_support=False)
    def search(self, query: str, extended: bool, exact: bool) -> Iterator[Package]:
        """Fetch matching packages.

        .. caution::
            Search does not support extended or exact matching. So we returns the best
            subset of results and let
            :py:meth:`meta_package_manager.base.PackageManager.refiltered_search` refine
            them.

        .. danger:
            `Cargo limits search to 100 results
            <https://doc.rust-lang.org/cargo/commands/cargo-search.html#search-options>`_,
            and because CLI output is refiltered as mentioned above, the final results
            can't be guaranteed.

        .. code-block:: shell-session

            $ cargo --color never --quiet search --limit 100 python
            python = "0.0.0"                  # Python.
            pyo3-asyncio = "0.16.0"           # PyO3 utilities for Python's Asyncio
            pyo3-asyncio-macros = "0.16.0"    # Proc Macro Attributes for PyO3 Asyncio
            pyo3 = "0.16.4"                   # Bindings to Python interpreter
            pyenv-python = "0.4.0"            # A pyenv shim for python
            python-launcher = "1.0.0"         # The Python launcher for Unix
            py-spy = "0.3.11"                 # Sampling profiler for Python programs
            python_mixin = "0.0.0"            # Use Python to generate your Rust, right…
            pyflow = "0.3.1"                  # A modern Python dependency manager
            pypackage = "0.0.3"               # A modern Python dependency manager
            ... and 1664 crates more (use --limit N to see more)
        """
        output = self.run_cli("search", "--limit", "100", query)

        for package_id, version, description in self._SEARCH_REGEXP.findall(output):
            yield self.package(
                id=package_id,
                description=description,
                latest_version=version,
            )

    @version_not_implemented
    def install(self, package_id: str, version: str | None = None) -> str:
        """Install one package.

        .. code-block:: shell-session

            $ cargo --color never install bore-cli
              Updating crates.io index
            Downloaded bore-cli v0.4.0
            Downloaded 1 crate (20.9 KB) in 0.26s
            Installing bore-cli v0.4.0
            Downloaded serde_derive v1.0.137
            Downloaded unicode-xid v0.2.3
            Downloaded clap_lex v0.2.0
            (...)
            Compiling bore-cli v0.4.0
              Finished release [optimized] target(s) in 1m 06s
             Replacing /home/mawoka/.cargo/bin/bore
              Replaced `bore-cli v0.2.3` with `bore-cli v0.4.0` (executable `bore`)
        """
        return self.run_cli("install", package_id)

    def remove(self, package_id: str) -> str:
        """Remove one package.

        .. code-block:: shell-session

            $ cargo --color never uninstall bore-cli
                Removing /Users/me/.cargo/bin/bore
        """
        return self.run_cli("uninstall", package_id)



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""Utilities to generate extra labels to use for GitHub issues and PRs."""

from __future__ import annotations

from boltons.iterutils import flatten
from extra_platforms import extract_members

from .inventory import MAIN_PLATFORMS
from .pool import pool

TYPE_CHECKING = False
if TYPE_CHECKING:
    TLabelSet = frozenset[str]
    TLabelGroup = dict[str, TLabelSet]


LABELS: list[tuple[str, str, str]] = [
    (
        "🔌 bar-plugin",
        "#fef2c0",
        "Xbar/SwiftBar plugin code, documentation and features",
    ),
]
"""Global registry of all labels used in the project.

Structure:

.. code-block:: python

    ("label_name", "color", "optional_description")
"""


def generate_labels(
    all_labels: TLabelSet,
    groups: TLabelGroup,
    prefix: str,
    color: str,
) -> dict[str, str]:
    """Generate labels.

    A dedicated label is produced for each entry of the ``all_labels`` parameter,
    unless it is part of a ``group``. In which case a dedicated label for that group
    will be created.
    """
    # Check all labels to group are referenced in the full label set.
    grouped_labels = set(flatten(groups.values()))
    assert grouped_labels.issubset(all_labels)

    label_map = {}

    # Create a dedicated label for each non-grouped entry.
    standalone_labels = all_labels - grouped_labels
    for label_id in standalone_labels:
        label_name = f"{prefix}{label_id}"
        # Check the addition of the prefix does not collide with an existing label.
        assert label_name not in all_labels
        label_map[label_id] = label_name
        # Register label to the global registry.
        LABELS.append((label_name, color, label_id))

    # Create a dedicated label for each group.
    for group_id, label_ids in groups.items():
        label_name = f"{prefix}{group_id}"
        # Check the addition of the prefix does not collide with an existing label.
        assert label_name not in all_labels
        for label_id in label_ids:
            label_map[label_id] = label_name
        # Build a description that is less than 100 characters.
        description = ""
        truncation_mark = ", …"
        for item_id in sorted(label_ids, key=str.casefold):
            new_item = f", {item_id}" if description else item_id
            if len(description) + len(new_item) <= 100 - len(truncation_mark):
                description += new_item
            else:
                description += truncation_mark
                break
        # Register label to the global registry.
        LABELS.append((label_name, color, description))

    # Sort label_map by their name.
    return dict(sorted(label_map.items(), key=lambda i: str.casefold(i[1])))


MANAGER_PREFIX = "📦 manager: "

MANAGER_LABEL_GROUPS: TLabelGroup = {
    "dnf-based": frozenset({"dnf", "dnf5", "yum"}),
    "dpkg-based": frozenset({"apt", "apt-mint", "opkg"}),
    "homebrew": frozenset({"brew", "cask"}),
    "npm-based": frozenset({"npm", "yarn"}),
    "pacman-based": frozenset({"pacman", "pacaur", "paru", "yay"}),
    "pip-based": frozenset({"pip", "pipx"}),
    "uv-based": frozenset({"uv", "uvx"}),
    "vscode-based": frozenset({"vscode", "vscodium"}),
}
"""Managers sharing some origin or implementation are grouped together under the same
label."""

all_manager_label_ids = frozenset(set(pool.all_manager_ids) | {"mpm"})
"""Adds ``mpm`` as its own manager alongside all those implemented."""

# Check group IDs do not collide with original labels.
assert all_manager_label_ids.isdisjoint(MANAGER_LABEL_GROUPS.keys())

MANAGER_LABELS = generate_labels(
    all_manager_label_ids,
    MANAGER_LABEL_GROUPS,
    MANAGER_PREFIX,
    "#bfdadc",
)
"""Maps all manager IDs to their labels."""


PLATFORM_PREFIX = "🖥 platform: "

PLATFORM_LABEL_GROUPS: TLabelGroup = {}
for p_obj in MAIN_PLATFORMS:
    PLATFORM_LABEL_GROUPS[p_obj.name] = frozenset(
        p.name for p in extract_members(p_obj)
    )
"""Similar platforms are grouped together under the same label."""

all_platform_label_ids = frozenset(flatten(PLATFORM_LABEL_GROUPS.values()))

PLATFORM_LABELS = generate_labels(
    all_platform_label_ids,
    PLATFORM_LABEL_GROUPS,
    PLATFORM_PREFIX,
    "#bfd4f2",
)
"""Maps all platform names to their labels."""

# Force sorting of labels.
LABELS = sorted(LABELS, key=lambda i: str.casefold(i[0]))



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import annotations

import json

from extra_platforms import LINUX_LIKE

from ..base import PackageManager
from ..capabilities import version_not_implemented
from ..version import parse_version

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Iterator

    from ..base import Package


class FWUPD(PackageManager):
    name = "Linux Vendor Firmware Service"

    homepage_url = "https://fwupd.org"

    platforms = LINUX_LIKE

    requirement = "1.9.5"
    """Version `1.9.5 is the first supporting --json parameter for get-devices command
    <https://github.com/fwupd/fwupd/releases/tag/1.9.5>`_.
    """

    cli_names = ("fwupdmgr",)

    pre_args = (
        # Answer yes to all questions.
        "--assume-yes",
        # Do not check or prompt for reboot after update.
        "--no-reboot-check",
        # Do not prompt for devices.
        "--no-device-prompt",
    )

    version_regexes = (r"compile\s+org\.freedesktop\.fwupd\s+(?P<version>\S+)\s+",)
    """
    .. code-block:: shell-session

        $ fwupdmgr --version
        compile   com.hughsie.libxmlb           0.3.18
        compile   com.hughsie.libjcat           0.2.0
        compile   org.freedesktop.fwupd         1.9.24
        runtime   org.freedesktop.fwupd-efi     1.4
        compile   org.freedesktop.gusb          0.4.8
        runtime   com.hughsie.libxmlb           0.3.x
        runtime   org.freedesktop.gusb          0.4.8
        runtime   com.hughsie.libjcat           0.2.0
        runtime   org.freedesktop.fwupd         1.9.24
        runtime   org.kernel                    6.8.0-48-generic
    """

    @property
    def installed(self) -> Iterator[Package]:
        """Fetch installed packages.

        .. code-block:: shell-session

            $ fwupdmgr --assume-yes --no-reboot-check --no-device-prompt get-devices --json | jq
            {
              "Devices": [
                {
                  "Name": "USB2.0 Hub",
                  "DeviceId": "7622d5fdbf1d1e08138156da7d83bf693986ad16",
                  "ParentDeviceId" : "b5540761dfe33d9abccd3bb21f1d725f9e69f541",
                  "CompositeId" : "b5540761dfe33d9abccd3bb21f1d725f9e69f541",
                  "InstanceIds": [
                    "USB\\VID_17EF&PID_3080",
                    "USB\\VID_17EF&PID_3080&REV_5163",
                    "USB\\VID_17EF&PID_3080&HUB_20",
                    "USB\\VID_17EF&PID_3080&SPI_C220",
                    "USB\\VID_17EF&PID_3080&SPI_C220&REV_5163",
                    "USB\\VID_17EF&PID_3080&DEV_VL820Q7"
                  ],
                  "Guid": [
                    "8ee94f0e-9b44-596a-bdd9-6f90401664cc",
                    "35199e34-cf82-5b09-9287-622d225056e4",
                    "0987e3c9-b1ee-5763-ac6e-51329b034e4b",
                    "163cea66-5a78-58af-80ba-21be960aae5c",
                    "c7def18d-66ae-5531-924b-2020c3638181"
                  ],
                  "Summary": "USB 3.x hub",
                  "Plugin": "vli",
                  "Protocol" : "com.vli.usbhub",
                  "Flags": [
                    "updatable",
                    "registered",
                    "can-verify",
                    "can-verify-image",
                    "dual-image",
                    "self-recovery",
                    "add-counterpart-guids",
                    "unsigned-payload"
                  ],
                  "Vendor": "VIA Labs, Inc.",
                  "VendorId" : "USB:0x17EF",
                  "Version": "51.63",
                  "VersionFormat" : "bcd",
                  "VersionRaw": 20835,
                  "Icons": [
                    "usb-hub"
                  ],
                  "InstallDuration" : 15,
                  "Created": 1686048073
                },
                {
                  "DeviceId" : "20de1d77d0d1787bc56ef62f7d05de49361e1e07",
                  "InstanceIds" : [
                    "DRM\\VEN_RHT&DEV_1234"
                  ],
                  "Guid" : [
                    "90b1437c-86da-5374-a9a7-ceca8b0afd5e"
                  ],
                  "Plugin" : "linux_display",
                  "Flags" : [
                    "registered"
                  ],
                  "Created" : 1731659840
                },
                {
                  "Name" : "UEFI dbx",
                  "DeviceId" : "362301da643102b9f38477387e2193e57abaa590",
                  "InstanceIds" : [
                    "UEFI\\CRT_E1FFABB40A30D9EE750BDA8BAF36ACA304FF20526138129247576B3339C54537&ARCH_AA64",
                    "UEFI\\CRT_A1117F516A32CEFCBA3F2D1ACE10A87972FD6BBE8FE0D0B996E09E65D802A503&ARCH_AA64"
                  ],
                  "Guid" : [
                    "a9b31b16-b184-560f-97cb-1aa25e418c7d",
                    "67d35028-ca5b-5834-834a-f97380381082"
                  ],
                  "Summary" : "UEFI revocation database",
                  "Plugin" : "uefi_dbx",
                  "Protocol" : "org.uefi.dbx",
                  "Flags" : [
                    "internal",
                    "updatable",
                    "supported",
                    "registered",
                    "needs-reboot",
                    "usable-during-update",
                    "only-version-upgrade",
                    "signed-payload"
                  ],
                  "Checksums" : [
                    "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855"
                  ],
                  "VendorId" : "UEFI:Linux Foundation",
                  "Version" : "0",
                  "VersionLowest" : "0",
                  "VersionFormat" : "number",
                  "Icons" : [
                    "computer"
                  ],
                  "InstallDuration" : 1,
                  "Created" : 1731659840,
                  "Releases" : [
                    {
                      "AppstreamId" : "org.linuxfoundation.dbx.aa64.firmware",
                      "ReleaseId" : "35289",
                      "RemoteId" : "lvfs",
                      "Name" : "Secure Boot dbx",
                      "NameVariantSuffix" : "aa64",
                      "Summary" : "UEFI Secure Boot Forbidden Signature Database",
                      "Description" : "<p>Insecure versions of the Microsoft Windows boot manager affected by Black Lotus were added to the list of forbidden signatures due to a discovered security problem.This updates the dbx to the latest release from Microsoft.</p><p>Before installing the update, fwupd will check for any affected executables in the ESP and will refuse to update if it finds any boot binaries signed with any of the forbidden signatures.Applying this update may also cause some Windows install media to not start correctly.</p>",
                      "Version" : "26",
                      "Filename" : "DBXUpdate-20230509-aa64.cab",
                      "Protocol" : "org.uefi.dbx",
                      "Categories" : [
                        "X-Configuration",
                        "X-System"
                      ],
                      "Issues" : [
                        "CVE-2022-21894"
                      ],
                      "Checksum" : [
                        "46a42362cd34c0d103cf534ca431508d24715e51",
                        "3ff3f17a9e5d372e51503803f22294f32ca90d1fe570b0bef4088c3a542617e6"
                      ],
                      "License" : "LicenseRef-proprietary",
                      "Size" : 4610,
                      "Created" : 1683590400,
                      "Locations" : [
                        "https://fwupd.org/downloads/3ff3f17a9e5d372e51503803f22294f32ca90d1fe570b0bef4088c3a542617e6-DBXUpdate-20230509-aa64.cab"
                      ],
                      "Uri" : "https://fwupd.org/downloads/3ff3f17a9e5d372e51503803f22294f32ca90d1fe570b0bef4088c3a542617e6-DBXUpdate-20230509-aa64.cab",
                      "Homepage" : "https://uefi.org/revocationlistfile",
                      "Vendor" : "Linux Foundation",
                      "Flags" : [
                        "trusted-metadata",
                        "is-upgrade"
                      ],
                      "InstallDuration" : 1
                    },
                    {
                      "AppstreamId" : "org.linuxfoundation.dbx.aa64.firmware",
                      "ReleaseId" : "28503",
                      "RemoteId" : "lvfs",
                      "Name" : "Secure Boot dbx",
                      "NameVariantSuffix" : "aa64",
                      "Summary" : "UEFI Secure Boot Forbidden Signature Database",
                      "Description" : "<p>An insecure version of software from vmware has been added to the list of forbidden signatures due to a discovered security problem.This updates the dbx to the latest release from Microsoft.</p><p>Before installing the update, fwupd will check for any affected executables in the ESP and will refuse to update if it finds any boot binaries signed with any of the forbidden signatures.</p>",
                      "Version" : "22",
                      "Filename" : "DBXUpdate-20230314-aa64.cab",
                      "Protocol" : "org.uefi.dbx",
                      "Categories" : [
                        "X-Configuration",
                        "X-System"
                      ],
                      "Issues" : [
                        "CVE-2023-28005"
                      ],
                      "Checksum" : [
                        "611e745638f05e9a11c2998cfba38f0bad651141",
                        "533ce4ac028585925268d9e39079b71730a7abd94f611bc532707938d4271ad3"
                      ],
                      "License" : "LicenseRef-proprietary",
                      "Size" : 4418,
                      "Created" : 1678752000,
                      "Locations" : [
                        "https://fwupd.org/downloads/533ce4ac028585925268d9e39079b71730a7abd94f611bc532707938d4271ad3-DBXUpdate-20230314-aa64.cab"
                      ],
                      "Uri" : "https://fwupd.org/downloads/533ce4ac028585925268d9e39079b71730a7abd94f611bc532707938d4271ad3-DBXUpdate-20230314-aa64.cab",
                      "Homepage" : "https://uefi.org/revocationlistfile",
                      "Vendor" : "Linux Foundation",
                      "Flags" : [
                        "trusted-metadata",
                        "is-upgrade"
                      ],
                      "InstallDuration" : 1
                    },
                    {
                      "AppstreamId" : "org.linuxfoundation.dbx.aa64.firmware",
                      "ReleaseId" : "15180",
                      "RemoteId" : "lvfs",
                      "Name" : "Secure Boot dbx",
                      "NameVariantSuffix" : "aa64",
                      "Summary" : "UEFI Secure Boot Forbidden Signature Database",
                      "Description" : "<p>This updates the dbx to the latest release from Microsoft which adds insecure versions of grub and shim to the list of forbidden signatures due to multiple discovered security updates.</p>",
                      "Version" : "21",
                      "Filename" : "DBXUpdate-20220812-aa64.cab",
                      "Protocol" : "org.uefi.dbx",
                      "Categories" : [
                        "X-Configuration",
                        "X-System"
                      ],
                      "Issues" : [
                        "CVE-2022-34303",
                        "309662",
                        "CVE-2022-34302",
                        "CVE-2022-34301"
                      ],
                      "Checksum" : [
                        "4032a1d8734e6085f4a6e4bb26a038eb639603b9",
                        "bf56092de6586604d2b41d5bb4c9b7787a07adde408fd4134a3f3606f7fda999"
                      ],
                      "License" : "LicenseRef-proprietary",
                      "Size" : 4370,
                      "Created" : 1595980800,
                      "Locations" : [
                        "https://fwupd.org/downloads/bf56092de6586604d2b41d5bb4c9b7787a07adde408fd4134a3f3606f7fda999-DBXUpdate-20220812-aa64.cab"
                      ],
                      "Uri" : "https://fwupd.org/downloads/bf56092de6586604d2b41d5bb4c9b7787a07adde408fd4134a3f3606f7fda999-DBXUpdate-20220812-aa64.cab",
                      "Homepage" : "https://uefi.org/revocationlistfile",
                      "Vendor" : "Linux Foundation",
                      "Flags" : [
                        "trusted-metadata",
                        "is-upgrade"
                      ],
                      "InstallDuration" : 1
                    }
                  ]
                },
                {
                  "Name" : "Virtio network device",
                  "DeviceId" : "17076870bcf7a84a9c8e999d7e54e39b446032bb",
                  "InstanceIds" : [
                    "PCI\\VEN_1AF4&DEV_1000",
                    "PCI\\VEN_1AF4&DEV_1000&SUBSYS_1AF40001"
                  ],
                  "Guid" : [
                    "21c85fac-5270-576f-a84e-04969f8cf75a",
                    "b93ef629-0df1-5505-9fee-6992b8b9abd8"
                  ],
                  "Plugin" : "optionrom",
                  "Flags" : [
                    "internal",
                    "registered",
                    "can-verify",
                    "can-verify-image"
                  ],
                  "Vendor" : "Red Hat, Inc.",
                  "VendorId" : "PCI:0x1AF4",
                  "Created" : 1731659840
                }
              ]
            }
        """
        output = self.run_cli("get-devices", "--json")

        if output:
            for device in json.loads(output)["Devices"]:
                if "updatable" in device["Flags"]:
                    yield self.package(
                        id=device["DeviceId"],
                        name=device["Name"],
                        installed_version=device["Version"],
                    )

    @property
    def outdated(self) -> Iterator[Package]:
        """Fetch outdated packages.

        .. code-block:: shell-session

            $ fwupdmgr --assume-yes --no-reboot-check --no-device-prompt get-updates --json | jq
            {
              "Devices" : [
                {
                  "Name" : "UEFI dbx",
                  "DeviceId" : "362301da643102b9f38477387e2193e57abaa590",
                  "InstanceIds" : [
                    "UEFI\\CRT_E1FFABB40A30D9EE750BDA8BAF36ACA304FF20526138129247576B3339C54537&ARCH_AA64",
                    "UEFI\\CRT_A1117F516A32CEFCBA3F2D1ACE10A87972FD6BBE8FE0D0B996E09E65D802A503&ARCH_AA64"
                  ],
                  "Guid" : [
                    "a9b31b16-b184-560f-97cb-1aa25e418c7d",
                    "67d35028-ca5b-5834-834a-f97380381082"
                  ],
                  "Summary" : "UEFI revocation database",
                  "Plugin" : "uefi_dbx",
                  "Protocol" : "org.uefi.dbx",
                  "Flags" : [
                    "internal",
                    "updatable",
                    "supported",
                    "registered",
                    "needs-reboot",
                    "usable-during-update",
                    "only-version-upgrade",
                    "signed-payload"
                  ],
                  "Checksums" : [
                    "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855"
                  ],
                  "VendorId" : "UEFI:Linux Foundation",
                  "Version" : "0",
                  "VersionLowest" : "0",
                  "VersionFormat" : "number",
                  "Icons" : [
                    "computer"
                  ],
                  "InstallDuration" : 1,
                  "Created" : 1731659840,
                  "Releases" : [
                    {
                      "AppstreamId" : "org.linuxfoundation.dbx.aa64.firmware",
                      "ReleaseId" : "35289",
                      "RemoteId" : "lvfs",
                      "Name" : "Secure Boot dbx",
                      "NameVariantSuffix" : "aa64",
                      "Summary" : "UEFI Secure Boot Forbidden Signature Database",
                      "Description" : "<p>Insecure versions of the Microsoft Windows boot manager affected by Black Lotus were added to the list of forbidden signatures due to a discovered security problem.This updates the dbx to the latest release from Microsoft.</p><p>Before installing the update, fwupd will check for any affected executables in the ESP and will refuse to update if it finds any boot binaries signed with any of the forbidden signatures.Applying this update may also cause some Windows install media to not start correctly.</p>",
                      "Version" : "26",
                      "Filename" : "DBXUpdate-20230509-aa64.cab",
                      "Protocol" : "org.uefi.dbx",
                      "Categories" : [
                        "X-Configuration",
                        "X-System"
                      ],
                      "Issues" : [
                        "CVE-2022-21894"
                      ],
                      "Checksum" : [
                        "46a42362cd34c0d103cf534ca431508d24715e51",
                        "3ff3f17a9e5d372e51503803f22294f32ca90d1fe570b0bef4088c3a542617e6"
                      ],
                      "License" : "LicenseRef-proprietary",
                      "Size" : 4610,
                      "Created" : 1683590400,
                      "Locations" : [
                        "https://fwupd.org/downloads/3ff3f17a9e5d372e51503803f22294f32ca90d1fe570b0bef4088c3a542617e6-DBXUpdate-20230509-aa64.cab"
                      ],
                      "Uri" : "https://fwupd.org/downloads/3ff3f17a9e5d372e51503803f22294f32ca90d1fe570b0bef4088c3a542617e6-DBXUpdate-20230509-aa64.cab",
                      "Homepage" : "https://uefi.org/revocationlistfile",
                      "Vendor" : "Linux Foundation",
                      "Flags" : [
                        "trusted-metadata",
                        "is-upgrade"
                      ],
                      "InstallDuration" : 1
                    },
                    {
                      "AppstreamId" : "org.linuxfoundation.dbx.aa64.firmware",
                      "ReleaseId" : "28503",
                      "RemoteId" : "lvfs",
                      "Name" : "Secure Boot dbx",
                      "NameVariantSuffix" : "aa64",
                      "Summary" : "UEFI Secure Boot Forbidden Signature Database",
                      "Description" : "<p>An insecure version of software from vmware has been added to the list of forbidden signatures due to a discovered security problem.This updates the dbx to the latest release from Microsoft.</p><p>Before installing the update, fwupd will check for any affected executables in the ESP and will refuse to update if it finds any boot binaries signed with any of the forbidden signatures.</p>",
                      "Version" : "22",
                      "Filename" : "DBXUpdate-20230314-aa64.cab",
                      "Protocol" : "org.uefi.dbx",
                      "Categories" : [
                        "X-Configuration",
                        "X-System"
                      ],
                      "Issues" : [
                        "CVE-2023-28005"
                      ],
                      "Checksum" : [
                        "611e745638f05e9a11c2998cfba38f0bad651141",
                        "533ce4ac028585925268d9e39079b71730a7abd94f611bc532707938d4271ad3"
                      ],
                      "License" : "LicenseRef-proprietary",
                      "Size" : 4418,
                      "Created" : 1678752000,
                      "Locations" : [
                        "https://fwupd.org/downloads/533ce4ac028585925268d9e39079b71730a7abd94f611bc532707938d4271ad3-DBXUpdate-20230314-aa64.cab"
                      ],
                      "Uri" : "https://fwupd.org/downloads/533ce4ac028585925268d9e39079b71730a7abd94f611bc532707938d4271ad3-DBXUpdate-20230314-aa64.cab",
                      "Homepage" : "https://uefi.org/revocationlistfile",
                      "Vendor" : "Linux Foundation",
                      "Flags" : [
                        "trusted-metadata",
                        "is-upgrade"
                      ],
                      "InstallDuration" : 1
                    },
                    {
                      "AppstreamId" : "org.linuxfoundation.dbx.aa64.firmware",
                      "ReleaseId" : "15180",
                      "RemoteId" : "lvfs",
                      "Name" : "Secure Boot dbx",
                      "NameVariantSuffix" : "aa64",
                      "Summary" : "UEFI Secure Boot Forbidden Signature Database",
                      "Description" : "<p>This updates the dbx to the latest release from Microsoft which adds insecure versions of grub and shim to the list of forbidden signatures due to multiple discovered security updates.</p>",
                      "Version" : "21",
                      "Filename" : "DBXUpdate-20220812-aa64.cab",
                      "Protocol" : "org.uefi.dbx",
                      "Categories" : [
                        "X-Configuration",
                        "X-System"
                      ],
                      "Issues" : [
                        "CVE-2022-34303",
                        "309662",
                        "CVE-2022-34302",
                        "CVE-2022-34301"
                      ],
                      "Checksum" : [
                        "4032a1d8734e6085f4a6e4bb26a038eb639603b9",
                        "bf56092de6586604d2b41d5bb4c9b7787a07adde408fd4134a3f3606f7fda999"
                      ],
                      "License" : "LicenseRef-proprietary",
                      "Size" : 4370,
                      "Created" : 1595980800,
                      "Locations" : [
                        "https://fwupd.org/downloads/bf56092de6586604d2b41d5bb4c9b7787a07adde408fd4134a3f3606f7fda999-DBXUpdate-20220812-aa64.cab"
                      ],
                      "Uri" : "https://fwupd.org/downloads/bf56092de6586604d2b41d5bb4c9b7787a07adde408fd4134a3f3606f7fda999-DBXUpdate-20220812-aa64.cab",
                      "Homepage" : "https://uefi.org/revocationlistfile",
                      "Vendor" : "Linux Foundation",
                      "Flags" : [
                        "trusted-metadata",
                        "is-upgrade"
                      ],
                      "InstallDuration" : 1
                    }
                  ]
                }
              ]
            }
        """
        output = self.run_cli("get-updates", "--json")

        if output:
            for device in json.loads(output)["Devices"]:
                if "updatable" in device["Flags"] and device.get("Releases"):
                    yield self.package(
                        id=device["DeviceId"],
                        name=device["Name"],
                        latest_version=max(
                            parse_version(rel["Version"])
                            for rel in device.get("Releases")
                        ),
                        installed_version=device["Version"],
                    )

    def install(self, package_id: str, version: str | None = None) -> str:
        """Install one package.

        .. code-block:: shell-session

            $ sudo fwupdmgr  --assume-yes --no-reboot-check --no-device-prompt install 362301da643102b9f38477387e2193e57abaa590
            WARNING: UEFI capsule updates not available or enabled in firmware setup
            See https://github.com/fwupd/fwupd/wiki/PluginFlag:capsules-unsupported for more information.
            0.	Cancel
            1.	26
            2.	22
            3.	21
            Choose release [0-3]: 3
            Scheduling…              [***************************************]
            Successfully installed firmware

            $ sudo fwupdmgr --assume-yes --no-reboot-check --no-device-prompt install 362301da643102b9f38477387e2193e57abaa590 21
            WARNING: UEFI capsule updates not available or enabled in firmware setup
            See https://github.com/fwupd/fwupd/wiki/PluginFlag:capsules-unsupported for more information.
            Scheduling…              [***************************************]
            362301da643102b9f38477387e2193e57abaa590 is already scheduled to be updated
        """
        package_specs = package_id
        if version:
            package_specs += f" {version}"
        return self.run_cli("install", package_specs, sudo=True)

    def upgrade_all_cli(self) -> tuple[str, ...]:
        """Generates the CLI to upgrade all packages (default) or only the one provided
        as parameter.

        .. code-block:: shell-session

            $ sudo fwupdmgr --assume-yes --no-reboot-check --no-device-prompt update
        """
        return self.build_cli("update", sudo=True)

    @version_not_implemented
    def upgrade_one_cli(
        self,
        package_id: str,
        version: str | None = None,
    ) -> tuple[str, ...]:
        """Generates the CLI to upgrade the package provided as parameter.

        .. code-block:: shell-session

            $ sudo fwupdmgr --assume-yes --no-reboot-check --no-device-prompt update 362301da643102b9f38477387e2193e57abaa590
            WARNING: UEFI capsule updates not available or enabled in firmware setup
            See https://github.com/fwupd/fwupd/wiki/PluginFlag:capsules-unsupported for more information.
            Scheduling…              [ -                                     ]
            362301da643102b9f38477387e2193e57abaa590 is already scheduled to be updated
        """
        return self.build_cli("update", package_id, sudo=True)

    def sync(self) -> None:
        """Sync package metadata.

        .. code-block:: shell-session

            $ fwupdmgr --assume-yes --no-reboot-check --no-device-prompt refresh --force
            Updating lvfs
            Downloading…             [***************************************]
            Successfully downloaded new metadata: 1 local device supported
        """
        self.run_cli("refresh", "--force")



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import annotations

import re

from extra_platforms import UNIX_WITHOUT_MACOS

from ..base import PackageManager
from ..capabilities import search_capabilities, version_not_implemented

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Iterator

    from ..base import Package


class APT(PackageManager):
    """Base package manager shared by variation of the apt command.

    Documentation:
    - https://wiki.debian.org/AptCLI
    - http://manpages.ubuntu.com/manpages/xenial/man8/apt.8.html

    See other command equivalences at: https://wiki.archlinux.org/title/Pacman/Rosetta
    """

    homepage_url = "https://wiki.debian.org/AptCLI"

    platforms = UNIX_WITHOUT_MACOS

    requirement = "1.0.0"

    pre_args = ("--quiet",)

    _INSTALLED_REGEXP = re.compile(r"(\S+)\/\S+ (\S+) (\S+) .*")
    _OUTDATED_REGEXP = re.compile(r"(\S+)\/\S+ (\S+).*\[upgradable from: (\S+)\]")
    _SEARCH_REGEXP = re.compile(
        r"""
        ^(?P<package_id>\S+)  # A string with a char at least.
        /\S+\                 # A slash, any non-spaced string, then a space.
        (?P<version>\S+)      # Any non-spaced string.
        \                     # A space.
        (?:.+)\n              # Any content ending the line.
        (?P<description>      # Start of the multi-line desc group.
            (?:\ \ .+\n?)+    # Line(s) of content prefixed by 2 spaces.
        )
        """,
        re.MULTILINE | re.VERBOSE,
    )
    """
    ``--quiet``: produces output suitable for logging, omitting progress indicators.

    Source: https://manpages.org/apt-get/8#options
    """

    version_regexes = (r"apt\s+(?P<version>\S+)",)
    """
    .. code-block:: shell-session

        $ apt --version
        apt 2.0.6 (amd64)
    """

    @property
    def installed(self) -> Iterator[Package]:
        """Fetch installed packages.

        .. code-block:: shell-session

            $ apt --quiet list --installed
            Listing...
            adduser/xenial,now 3.113+nmu3ubuntu4 all [installed]
            bc/xenial,now 1.06.95-9build1 amd64 [installed]
            bsdmainutils/xenial,now 9.0.6ubuntu3 amd64 [installed,automatic]
            ca-certificates/xenial,now 20160104ubuntu1 all [installed]
            cron/xenial,now 3.0pl1-128ubuntu2 amd64 [installed]
            debconf/xenial,now 1.5.58ubuntu1 all [installed]
            debianutils/xenial,now 4.7 amd64 [installed]
            diffutils/xenial,now 1:3.3-3 amd64 [installed]
            e2fsprogs/xenial,now 1.42.13-1ubuntu1 amd64 [installed]
            ethstatus/xenial,now 0.4.3ubuntu2 amd64 [installed]
            file/xenial,now 1:5.25-2ubuntu1 amd64 [installed]
            findutils/xenial,now 4.6.0+git+20160126-2 amd64 [installed]
            libidn2-0/jammy,now 2.3.2-2build1 amd64 [installed,automatic]
            libidn2-0/jammy,now 2.3.2-2build1 i386 [installed,automatic]
        """
        output = self.run_cli("list", "--installed")

        for package in output.splitlines():
            match = self._INSTALLED_REGEXP.match(package)
            if match:
                package_id, installed_version, arch = match.groups()
                yield self.package(
                    id=package_id, installed_version=installed_version, arch=arch
                )

    @property
    def outdated(self) -> Iterator[Package]:
        """Fetch outdated packages.

        .. code-block:: shell-session

            $ apt --quiet list --upgradable
            Listing...
            apt/xenial-updates 1.2.19 amd64 [upgradable from: 1.2.15ubuntu0.2]
            nano/xenial-updates 2.5.3-2ubuntu2 amd64 [upgradable from: 2.5.3-2]
        """
        output = self.run_cli("list", "--upgradable")

        for package in output.splitlines():
            match = self._OUTDATED_REGEXP.match(package)
            if match:
                package_id, latest_version, installed_version = match.groups()
                yield self.package(
                    id=package_id,
                    latest_version=latest_version,
                    installed_version=installed_version,
                )

    def search(self, query: str, extended: bool, exact: bool) -> Iterator[Package]:
        """Fetch matching packages.

        .. code-block:: shell-session

            $ apt --quiet search abc --names-only
            Sorting...
            Full Text Search...
            abcde/xenial 2.7.1-1 all
              A Better CD Encoder

            abcmidi/xenial 20160103-1 amd64
              converter from ABC to MIDI format and back

            berkeley-abc/xenial 1.01+20150706hgc3698e0+dfsg-2 amd64
              ABC - A System for Sequential Synthesis and Verification

            fuse-overlayfs/jammy,now 1.7.1-1 amd64 [installed]
              implementation of overlay+shiftfs in FUSE for rootless containers

            grabcd-rip/xenial 0009-1 all
              rip and encode audio CDs - ripper

            libakonadi-kabc4/xenial 4:4.14.10-1ubuntu2 amd64
              Akonadi address book access library

        .. code-block:: shell-session

            $ apt --quiet search ^sed$ --names-only
            Sorting...
            Full Text Search...
            sed/xenial 2.1.9-3 all
              Blah blah blah

        .. code-block:: shell-session

            $ apt --quiet search abc --full
            Sorting...
            Full Text Search...
            abcde/xenial 2.7.1-1 all
              This package contains the essential basic system utilities.
              .
              Specifically, this package includes:
              basename cat chgrp chmod chown chroot cksum comm cp csplit cut
              dircolors dirname du echo env expand expr factor false fmt
              hostid id install join link ln logname ls md5sum mkdir mkfifo
              nohup od paste pathchk pinky pr printenv printf ptx pwd
              sha1sum seq shred sleep sort split stat stty sum sync tac tail
              tr true tsort tty uname unexpand uniq unlink users vdir wc who

            (...)

            midi/xenial 20160103-1 amd64
              converter from ABC to MIDI format and back
        """
        search_arg = "--names-only"
        if exact:
            # Rely on apt regexp support to speed-up exact match.
            query = f"^{query}$"
        # Extended search are always non-exact.
        elif extended:
            # Include full description in extended search to check up the match
            # in the CLI output after its execution.
            search_arg = "--full"

        output = self.run_cli("search", query, search_arg)

        for package_id, version, description in self._SEARCH_REGEXP.findall(output):
            yield self.package(
                id=package_id,
                description=description,
                latest_version=version,
            )

    @version_not_implemented
    def install(self, package_id: str, version: str | None = None) -> str:
        """Install one package.

        .. code-block:: shell-session

            $ sudo apt --quiet --yes install git
        """
        return self.run_cli("--yes", "install", package_id, sudo=True)

    def upgrade_all_cli(self) -> tuple[str, ...]:
        """Generates the CLI to upgrade all packages (default) or only the one provided
        as parameter.

        .. code-block:: shell-session

            $ sudo apt --quiet --yes upgrade
        """
        return self.build_cli("--yes", "upgrade", sudo=True)

    @version_not_implemented
    def upgrade_one_cli(
        self,
        package_id: str,
        version: str | None = None,
    ) -> tuple[str, ...]:
        """Generates the CLI to upgrade all packages (default) or only the one provided
        as parameter.

        .. code-block:: shell-session

            $ sudo apt --quiet --yes install --only-upgrade git
        """
        return self.build_cli(
            "--yes",
            "install",
            "--only-upgrade",
            package_id,
            sudo=True,
        )

    def sync(self) -> None:
        """Sync package metadata.

        .. code-block:: shell-session

            $ sudo apt --quiet --yes update
            Hit:1 http://archive.ubuntu.com xenial InRelease
            Get:2 http://archive.ubuntu.com xenial-updates InRelease [102 kB]
            Get:3 http://archive.ubuntu.com xenial-security InRelease [102 kB]
            Get:4 http://archive.ubuntu.com xenial/main Translation-en [568 kB]
            Fetched 6,868 kB in 2s (2,680 kB/s)
            Reading package lists...
            Building dependency tree...
            Reading state information...
        """
        self.run_cli("--yes", "update", sudo=True)

    def cleanup(self) -> None:
        """Removes things we don't need anymore.

        .. code-block:: shell-session

            $ sudo apt --quiet --yes autoremove
            $ sudo apt --quiet --yes clean
        """
        for command in ("autoremove", "clean"):
            self.run_cli("--yes", command, sudo=True)


class APT_Mint(APT):
    """Special version of apt for Linux Mint.

    Exactly the same as its parent but implement specific version extraction.
    """

    name = "Linux Mint's apt"

    homepage_url = "https://github.com/kdeldycke/meta-package-manager/issues/52"

    _MINT_SEARCH_REGEXP = re.compile(
        r"""
        \S                       # One non-space character.
        \s+                      # One space or more.
        (?P<package_id>[^\s:]+)  # Any non-space until whitespace or semi-colon.
        (?:\:\S+)?               # Optional arch suffix after package and semi-colon
        \s+                      # One space or more.
        -                        # A dash.
        \ ?                      # An optional space.
        (?P<description>.+)?     # Optional content string.
        """,
        re.VERBOSE,
    )

    cli_names = ("apt",)

    version_cli_options = ("version", "apt")
    """
    .. code-block:: shell-session

        $ apt version apt
        1.6.11
    """

    @search_capabilities(extended_support=False)
    def search(self, query: str, extended: bool, exact: bool) -> Iterator[Package]:
        """Fetch matching packages.

        .. caution::
            Search does not supports extended matching.

        .. code-block:: shell-session

            $ /usr/local/bin/apt --quiet search sed
            v   librust-slog-2.5+erased-serde-dev  -
            p   python3-blessed                    - Practical wrapper
            i   sed                                - GNU stream editor
            p   sed:i386                           - GNU stream editor

        .. code-block:: shell-session

            $ /usr/local/bin/apt --quiet search ^sed$
            i   sed              - GNU stream editor
            p   sed:i386         - GNU stream editor
        """
        if exact:
            # Rely on apt regexp support to speed-up exact match.
            query = f"^{query}$"

        output = self.run_cli("search", query)

        for package_id, description in self._MINT_SEARCH_REGEXP.findall(output):
            yield self.package(id=package_id, description=description)



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import annotations

from extra_platforms import ALL_PLATFORMS

from ..base import PackageManager
from ..capabilities import version_not_implemented

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Iterator

    from ..base import Package


class VSCode(PackageManager):
    name = "Visual Studio Code"

    homepage_url = "https://code.visualstudio.com"

    platforms = ALL_PLATFORMS

    requirement = "1.60.0"

    cli_names = ("code",)
    """
    .. code-block:: shell-session

        $ code --version
        1.60.2
        7f6ab5485bbc008386c4386d08766667e155244e
        x64
    """

    @property
    def installed(self) -> Iterator[Package]:
        """Fetch installed packages.

        .. code-block:: shell-session

            $ code --list-extensions --show-versions
            ms-python.python@2021.9.1246542782
            ms-python.vscode-pylance@2021.9.3
            ms-toolsai.jupyter@2021.8.2041215044
            ms-toolsai.jupyter-keymap@1.0.0
            samuelcolvin.jinjahtml@0.16.0
            tamasfe.even-better-toml@0.14.2
            trond-snekvik.simple-rst@1.5.0
        """
        output = self.run_cli("--list-extensions", "--show-versions")

        for package in output.splitlines():
            package_id, installed_version = package.split("@")
            yield self.package(id=package_id, installed_version=installed_version)

    @version_not_implemented
    def install(self, package_id: str, version: str | None = None) -> str:
        """Install one package.

        .. code-block:: shell-session

            $ code --install-extension tamasfe.even-better-toml
        """
        return self.run_cli("--install-extension", package_id)

    def remove(self, package_id: str) -> str:
        """Remove one package.

        .. code-block:: shell-session

            $ code --uninstall-extension tamasfe.even-better-toml
        """
        return self.run_cli("--uninstall-extension", package_id)


class VSCodium(VSCode):
    homepage_url = "https://vscodium.com"
    cli_names = ("codium",)



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import annotations

import re

from extra_platforms import UNIX_WITHOUT_MACOS

from ..base import PackageManager
from ..capabilities import search_capabilities, version_not_implemented

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Iterator

    from ..base import Package


class Pacman(PackageManager):
    """See command equivalences at: https://wiki.archlinux.org/title/Pacman/Rosetta."""

    homepage_url = "https://wiki.archlinux.org/title/pacman"

    platforms = UNIX_WITHOUT_MACOS

    requirement = "5.0.0"

    pre_args = ("--noconfirm", "--color", "never")

    _INSTALLED_REGEXP = re.compile(r"(\S+) (\S+)")
    _OUTDATED_REGEXP = re.compile(r"(\S+) (\S+) -> (\S+)")
    _SEARCH_REGEXP = re.compile(
        r"(?P<repo_id>\S+?)/(?P<package_id>\S+)\s+(?P<version>\S+).*\n\s+(?P<description>.+)",
        re.MULTILINE | re.VERBOSE,
    )

    version_regexes = (r".*Pacman\s+v(?P<version>\S+)",)
    r"""Search version right after the ``Pacman `` string.

    .. code-block:: shell-session

        $ pacman --version

         .--.                  Pacman v6.0.1 - libalpm v13.0.1
        / _.-' .-.  .-.  .-.   Copyright (C) 2006-2021 Pacman Development Team
        \  '-. '-'  '-'  '-'   Copyright (C) 2002-2006 Judd Vinet
         '--'
                            This program may be freely redistributed under
                            the terms of the GNU General Public License.
    """

    @property
    def installed(self) -> Iterator[Package]:
        """Fetch installed packages.

        .. code-block:: shell-session

            $ pacman --noconfirm --query
            a52dec 0.7.4-11
            aalib 1.4rc5-14
            abseil-cpp 20211102.0-2
            accountsservice 22.08.8-2
            acl 2.3.1-2
            acme.sh 3.0.2-1
            acpi 1.7-3
            acpid 2.0.33-1
        """
        output = self.run_cli("--query")

        for package in output.splitlines():
            match = self._INSTALLED_REGEXP.match(package)
            if match:
                package_id, installed_version = match.groups()
                yield self.package(id=package_id, installed_version=installed_version)

    @property
    def outdated(self) -> Iterator[Package]:
        """Fetch outdated packages.

        .. code-block:: shell-session

            $ pacman --noconfirm --query --upgrades
            linux 4.19.1.arch1-1 -> 4.19.2.arch1-1
            linux-headers 4.19.1.arch1-1 -> 4.19.2.arch1-1
        """
        output = self.run_cli("--query", "--upgrades")

        for package in output.splitlines():
            match = self._OUTDATED_REGEXP.match(package)
            if match:
                package_id, installed_version, latest_version = match.groups()
                yield self.package(
                    id=package_id,
                    latest_version=latest_version,
                    installed_version=installed_version,
                )

    @search_capabilities(extended_support=False)
    def search(self, query: str, extended: bool, exact: bool) -> Iterator[Package]:
        """Fetch matching packages.

        .. caution::
            Search does not supports extended matching.

        .. code-block:: shell-session

            $ pacman --noconfirm --sync --search fire
            extra/dump_syms 0.0.7-1
                Symbol dumper for Firefox
            extra/firefox 99.0-1
                Standalone web browser from mozilla.org
            extra/firefox-i18n-ach 99.0-1
                Acholi language pack for Firefox
            extra/firefox-i18n-af 99.0-1
                Afrikaans language pack for Firefox
            extra/firefox-i18n-an 99.0-1
                Aragonese language pack for Firefox
            extra/firefox-i18n-ar 99.0-1
                Arabic language pack for Firefox
            extra/firefox-i18n-ast 99.0-1
                Asturian language pack for Firefox
        """
        if exact:
            query = f"^{query}$"

        output = self.run_cli("--sync", "--search", query)

        for _repo_id, package_id, version, description in self._SEARCH_REGEXP.findall(
            output,
        ):
            yield self.package(
                id=package_id,
                description=description,
                latest_version=version,
            )

    @version_not_implemented
    def install(self, package_id: str, version: str | None = None) -> str:
        """Install one package.

        .. code-block:: shell-session

            $ sudo pacman --noconfirm --sync firefox
        """
        return self.run_cli("--sync", package_id, sudo=True)

    def upgrade_all_cli(self) -> tuple[str, ...]:
        """Generates the CLI to upgrade the package provided as parameter.

        .. code-block:: shell-session

            $ sudo pacman --noconfirm --sync --refresh --sysupgrade
        """
        return self.build_cli("--sync", "--refresh", "--sysupgrade", sudo=True)

    @version_not_implemented
    def upgrade_one_cli(
        self,
        package_id: str,
        version: str | None = None,
    ) -> tuple[str, ...]:
        """Generates the CLI to upgrade the package provided as parameter.

        .. code-block:: shell-session

            $ sudo pacman --noconfirm --sync firefox
        """
        return self.build_cli("--sync", package_id, sudo=True)

    def remove(self, package_id: str) -> str:
        """Removes a package.

        .. code-block:: shell-session

            $ sudo pacman --noconfirm --remove firefox
        """
        return self.run_cli("--remove", package_id, sudo=True)

    def sync(self) -> None:
        """Sync package metadata.

        .. code-block:: shell-session

            $ pacman --noconfirm --sync --refresh
        """
        self.run_cli("--sync", "--refresh")

    def cleanup(self) -> None:
        """Removes things we don't need anymore.

        .. code-block:: shell-session

            $ sudo pacman --noconfirm --sync --clean --clean
        """
        self.run_cli("--sync", "--clean", "--clean", sudo=True)


class Pacaur(Pacman):
    """``Pacaur`` wraps ``pacman`` and shadows its options."""

    homepage_url = "https://github.com/E5ten/pacaur"

    requirement = "4.0.0"

    version_regexes = (r"pacaur\s+(?P<version>\S+)",)
    r"""Search version right after the ``pacaur`` string.

    .. code-block:: shell-session

        $ pacaur --version
        pacaur 4.8.6
    """


class Paru(Pacman):
    """``paru`` wraps ``pacman`` and shadows its options."""

    homepage_url = "https://github.com/Morganamilo/paru"

    # v1.9.3 is the first version implementing the --sysupgrade option.
    requirement = "1.9.3"

    version_regexes = (r"paru\s+v(?P<version>\S+)",)
    r"""Search version right after the ``paru`` string.

    .. code-block:: shell-session

        $ paru --version
        paru v1.10.0 - libalpm v13.0.1
    """


class Yay(Pacman):
    """``yay`` wraps ``pacman`` and shadows its options."""

    homepage_url = "https://github.com/Jguer/yay"

    requirement = "11.0.0"

    version_regexes = (r"yay\s+v(?P<version>\S+)",)
    r"""Search version right after the ``yay`` string.

    .. code-block:: shell-session

        $ yay --version
        yay v11.1.2 - libalpm v13.0.1
    """



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import annotations

import logging
import sys
from collections import Counter, namedtuple
from collections.abc import Iterable
from configparser import RawConfigParser
from datetime import datetime, timezone
from functools import partial
from io import TextIOWrapper
from operator import attrgetter
from pathlib import Path
from textwrap import dedent
from unittest.mock import patch

import tomli_w
from boltons.cacheutils import LRI, cached
from click_extra import (
    STRING,
    Choice,
    EnumChoice,
    File,
    IntRange,
    Section,
    argument,
    echo,
    file_path,
    group,
    option,
    option_group,
    pass_context,
)
from click_extra.colorize import KO, OK, default_theme as theme, highlight
from click_extra.commands import default_extra_params
from click_extra.table import TableFormat, TableFormatOption
from extra_platforms import reduce

from . import __version__, bar_plugin
from .base import (
    CLIError,
    Operations,
    PackageManager,
    highlight_cli_name,
    packages_asdict,
)
from .inventory import MAIN_PLATFORMS
from .output import (
    BarPluginRenderer,
    SortableField,
    SortedTableFormatOption,
    colored_diff,
    print_json,
    print_stats,
)
from .pool import pool
from .sbom import SBOM, SPDX, CycloneDX, ExportFormat
from .specifier import VERSION_SEP, Solver, Specifier

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib  # type: ignore[import-not-found]

TYPE_CHECKING = False
if TYPE_CHECKING:
    from typing import IO

    from click_extra import Context, Parameter


# Subcommand sections.
EXPLORE = Section("Explore subcommands")
MAINTENANCE = Section("Maintenance subcommands")
SNAPSHOTS = Section("Package snapshots subcommands")
SBOM_SECTION = Section("SBOM subcommands")


XKCD_MANAGER_ORDER = ("pip", "brew", "npm", "dnf", "apt", "steamcmd")
"""Sequence of package managers as defined by `XKCD #1654: Universal Install Script
<https://xkcd.com/1654/>`_.

See the corresponding :issue:`implementation rationale in issue #10 <10>`.
"""


def is_stdout(filepath: Path) -> bool:
    """Check if a file path is set to stdout.

    Prevents the creation of a ``-`` file in the current directory.
    """
    return str(filepath) == "-"


def prep_path(filepath: Path) -> IO | None:
    """Prepare the output file parameter for Click's echo function."""
    if is_stdout(filepath):
        return None
    return filepath.open("w", encoding="UTF-8")


def update_manager_selection(
    ctx: Context, param: Parameter, value: str | Iterable[str] | bool | None
) -> None:
    """Update global selection list of managers in the context.

    Accumulate and merge all manager selectors to form the initial population enforced by the user.
    """
    # Option has not been called.
    if value is None:
        return

    # Use a list to keep the natural order of selection.
    to_add: list[str] = []
    # Use a set because removal takes precedence over addition: we don't care
    # about user's order.
    to_remove: set[str] = set()

    assert param.name

    # Add the value of --manager list.
    if param.name == "manager":
        if value:
            assert isinstance(value, Iterable)
            to_add.extend(value)

    # Add the value of --exclude list.
    elif param.name == "exclude":
        if value:
            assert isinstance(value, Iterable)
            to_remove.update(value)

    # Update the list of managers with the XKCD preset.
    elif param.name == "xkcd":
        if value:
            to_add.extend(XKCD_MANAGER_ORDER)

    # Update selection with single selectors.
    else:
        # Because the parameter's name is transformed into a Python identifier on
        # instantiation, we have to reverse the process to get our value.
        # Example: --apt-mint => apt_mint => apt-mint
        manager_id = param.name.removeprefix("no_").replace("_", "-")
        assert manager_id in pool.all_manager_ids, (
            f"unrecognized single manager selector {param.name!r}"
        )

        # Normalize the value to a boolean.
        if isinstance(value, str):
            value = RawConfigParser.BOOLEAN_STATES.get(value.lower(), value)
        assert value in (
            manager_id,
            True,
            False,
        ), f"unexpected value {value!r} for {param!r}"

        if param.name.startswith("no_") ^ (value is False):
            to_remove.add(manager_id)
        else:
            to_add.append(manager_id)

    logging.debug(f"Managers added by {param}: {to_add}")
    logging.debug(f"Managers removed by {param}: {to_remove}")

    # Initialize the shared context object to accumulate there the selection results.
    if ctx.obj is None:
        ctx.obj = {}
    if to_add:
        ctx.obj.setdefault("managers_to_add", []).extend(to_add)
    if to_remove:
        ctx.obj.setdefault("managers_to_remove", set()).update(to_remove)


def single_manager_selectors():
    """Dynamiccaly creates a dedicated flag selector alias for each manager."""
    single_flags = []
    single_no_flags = []
    for manager_id, manager in pool.items():
        # Parameters do not have a deprecated flag.
        # See: https://github.com/pallets/click/issues/2263
        deprecated_msg = "(Deprecated) " if manager.deprecated else ""
        single_flags.append(
            option(
                f"--{manager_id}",
                flag_value=manager_id,
                default=None,
                help=f"{deprecated_msg}Select {manager.name}.",
                expose_value=False,
                callback=update_manager_selection,
            )
        )
        single_no_flags.append(
            option(
                f"--no-{manager_id}",
                flag_value=manager_id,
                default=None,
                help=f"{deprecated_msg}Deselect {manager.name}.",
                expose_value=False,
                callback=update_manager_selection,
            )
        )
    return *single_flags, *single_no_flags


def bar_plugin_path(ctx: Context, param: Parameter, value: str | None):
    """Print the location of the :doc:`Xbar/SwiftBar plugin <bar-plugin>`.

    Returns the normalized path of the standalone `bar_plugin.py
    <https://github.com/kdeldycke/meta-package-manager/meta_package_manager/bar_plugin.py>`_
    script that is distributed with this Python module. This
    is made available under the :option:`mpm --bar-plugin-path` option.

    Notice that the fully-qualified home directory get replaced by its
    shorthand (``~``) if applicable:

    - the full ``/home/user/.python/site-packages/mpm/bar_plugin.py`` path is
      simplified to ``~/.python/site-packages/mpm/bar_plugin.py``,
    - but ``/usr/bin/python3.10/mpm/bar_plugin.py`` is returned as-is.
    """
    # Option has not been called.
    if not value:
        return

    # Options is only available when CLI is installed from sources, not CLI is
    # a bundled executable.
    if "__compiled__" in globals():
        logging.debug("CLI running as a binary.")
        logging.critical(
            "Option --bar-plugin-path is only available for CLI installed from "
            "sources.",
        )
        ctx.exit(2)

    bar_path = Path(bar_plugin.__file__).expanduser().resolve()
    home_dir = Path.home()

    if bar_path.is_relative_to(home_dir):
        home_shorthand = Path("~")
        shorten_bar_path = home_shorthand / bar_path.relative_to(home_dir)
        assert shorten_bar_path.expanduser().resolve() == bar_path
        bar_path = shorten_bar_path
    echo(bar_path)
    ctx.exit()


# XXX Why is Sphinx skipping Click methods in documentation?

# Add a help subcommand:
# npm help <term>    search for help on <term>
# npm help npm       more involved overview


#    -j N, --jobs=N      Specify the allowed number of parallel C compiler
#                        jobs. Defaults to the system CPU count.


def custom_extra_params() -> list[Parameter]:
    """Replace the default ``TableFormatOption`` with our ``SortedTableFormatOption``."""
    params: list[Parameter] = []
    for param in default_extra_params():
        if isinstance(param, TableFormatOption):
            params.append(
                SortedTableFormatOption(
                    param_decls=("-o", "--output-format"),
                    type=EnumChoice(TableFormat),
                    help="Rendering format of the output.",
                )
            )
        else:
            params.append(param)
    return params


@group(
    # XXX Default verbosity has been changed in Click Extra 4.0.0 from INFO to WARNING.
    context_settings={"default_map": {"verbosity": "INFO"}},
    params=custom_extra_params(),
)
@option_group(
    "Package manager selection",
    # ---------------------- 80 characters reference limit ----------------------- #
    dedent("""\
    \b
    Use these options to restrict the subcommand to a subset of managers.

    \b
    - By default, mpm will evaluate all managers supported on the current platform.
    - Use the --<manager-id> selectors to restrict target to a subset of managers.
    - To remove a manager from the selection, use --no-<manager-id> selectors.
    - Order of the selectors is preserved for priority-sensitive subcommands.
    - Exclusion of a manager always takes precedence over its inclusion.
    \b

    """),
    *single_manager_selectors(),
    option(
        "-a",
        "--all-managers",
        is_flag=True,
        default=False,
        help="Force evaluation of all managers implemented by mpm, including those "
        "not supported by the current platform or deprecated. Still applies filtering "
        "by --<manager-id> / --no-<manager-id> options before calling the subcommand.",
    ),
    option(
        "-x",
        "--xkcd",
        is_flag=True,
        default=None,
        expose_value=False,
        callback=update_manager_selection,
        help="Preset manager selection as defined by XKCD #1654. Equivalent to: "
        "{}.".format(" ".join(f"--{mid}" for mid in XKCD_MANAGER_ORDER)),
    ),
    option(
        "-m",
        "--manager",
        type=Choice(pool.all_manager_ids, case_sensitive=False),
        multiple=True,
        default=None,
        expose_value=False,
        hidden=True,
        callback=update_manager_selection,
        help="(Deprecated) Use --<manager-id> single selector instead.",
    ),
    option(
        "-e",
        "--exclude",
        type=Choice(pool.all_manager_ids, case_sensitive=False),
        multiple=True,
        default=None,
        expose_value=False,
        hidden=True,
        callback=update_manager_selection,
        help="(Deprecated) Use --no-<manager-id> single selector instead.",
    ),
)
@option_group(
    "Manager options",
    option(
        "--ignore-auto-updates/--include-auto-updates",
        default=True,
        help="Report all outdated packages, including those tagged as "
        "auto-updating. Only applies to outdated and upgrade subcommands.",
    ),
    option(
        "--stop-on-error/--continue-on-error",
        default=False,
        help="Stop right away or continue operations on manager CLI error.",
    ),
    option(
        "-d",
        "--dry-run",
        is_flag=True,
        default=False,
        help="Do not actually perform any action, just simulate CLI calls.",
    ),
    option(
        "-t",
        "--timeout",
        type=IntRange(min=0),
        default=500,
        help="Set maximum duration in seconds for each CLI call.",
    ),
)
@option_group(
    "Output options",
    option(
        "--description",
        is_flag=True,
        default=False,
        help="Show package description in results.",
    ),
    option(
        "-s",
        "--sort-by",
        type=EnumChoice(SortableField),
        default=SortableField.MANAGER_ID,
        help="Sort results.",
    ),
    # option('--sort-asc/--sort-desc', default=True)
    option(
        "--stats/--no-stats",
        default=True,
        help="Print per-manager package statistics.",
    ),
)
@option_group(
    "Xbar/SwiftBar options",
    option(
        "--bar-plugin-path",
        is_flag=True,
        default=False,
        expose_value=False,
        is_eager=True,
        callback=bar_plugin_path,
        help="Print location of the Xbar/SwiftBar plugin.",
    ),
)
@pass_context
def mpm(
    ctx,
    all_managers,
    ignore_auto_updates,
    stop_on_error,
    dry_run,
    timeout,
    description,
    sort_by,
    stats,
):
    """CLI options shared by all subcommands."""
    # Silence all log messages for JSON rendering unless in debug mode.
    if (
        ctx.meta["click_extra.table_format"] == TableFormat.JSON
        and ctx.meta["click_extra.verbosity"] != "DEBUG"
    ):
        logging.disable()

        def remove_logging_override():
            """Reset the logging override to its default state.

            ``logging.disable()`` mess with the logging module internals at the root
            level. We need to restore the default behavior when the context is closed,
            otherwise the logging module will be stuck in a disabled state.

            See: https://docs.python.org/3/library/logging.html?highlight=logging#logging.disable
            """
            logging.disable(logging.NOTSET)

        ctx.call_on_close(remove_logging_override)

    # Normalize to None if no manager selectors have been used. This prevent the
    # pool.select_managers() method to iterate over an empty population of managers to
    # choose from.
    user_selection = None
    managers_to_remove = None
    if ctx.obj:
        user_selection = ctx.obj.get("managers_to_add", None)
        managers_to_remove = ctx.obj.get("managers_to_remove", None)
    selection_string = (
        " platform default"
        if not user_selection
        else "> " + " > ".join(map(theme.invoked_command, user_selection))
    )
    deselection_string = (
        "None"
        if not managers_to_remove
        else ", ".join(map(theme.invoked_command, sorted(managers_to_remove)))
    )
    logging.info(f"User selection of managers by priority:{selection_string}")
    logging.info(f"Managers dropped by user: {deselection_string}")

    # Select the subset of manager to target, and apply manager-level options.
    selected_managers = partial(
        pool.select_managers,
        keep=user_selection,
        drop=managers_to_remove,
        keep_deprecated=all_managers,
        # Should we include auto-update packages or not?
        ignore_auto_updates=ignore_auto_updates,
        # Does the manager should raise on error or not.
        stop_on_error=stop_on_error,
        dry_run=dry_run,
        timeout=timeout,
    )

    # Load up current and new global options to the context for subcommand consumption.
    ctx.obj = namedtuple(
        "GlobalOptions",
        (
            "selected_managers",
            "description",
            "sort_by",
            "stats",
        ),
        defaults=(
            selected_managers,
            description,
            sort_by,
            stats,
        ),
    )()


@mpm.command(
    short_help="List supported package managers and their location.",
    section=EXPLORE,
)
@pass_context
def managers(ctx):
    """List all supported package managers and autodetect their presence on the
    system."""
    select_params = {
        # Do not drop managers whose CLI was not found. Keep them to show off how
        # mpm is reacting to the local platform.
        "drop_not_found": False,
    }

    # Machine-friendly data rendering.
    if ctx.meta["click_extra.table_format"] == TableFormat.JSON:
        manager_data = {}
        # Build up the data structure of manager metadata.
        fields = (
            "name",
            "id",
            "supported",
            "cli_path",
            "executable",
            "version",
            "fresh",
            "available",
        )
        for manager in ctx.obj.selected_managers(**select_params):
            manager_data[manager.id] = {fid: getattr(manager, fid) for fid in fields}
            # Serialize errors at the last minute to gather all we encountered.
            manager_data[manager.id]["errors"] = list(
                {expt.error for expt in manager.cli_errors},
            )

        print_json(manager_data)
        ctx.exit()

    # Human-friendly content rendering.
    table = []
    for manager in ctx.obj.selected_managers(**select_params):
        # Build up the OS column content.
        os_infos = OK if manager.supported else KO
        if not manager.supported:
            os_infos += " {}".format(
                ", ".join(
                    sorted(p.name for p in reduce(manager.platforms, MAIN_PLATFORMS))
                ),
            )
        if manager.deprecated:
            os_infos += f" {theme.warning('(deprecated)')}"

        # Build up the CLI path column content.
        cli_infos = "{} {}".format(
            OK if manager.cli_path else KO,
            highlight_cli_name(manager.cli_path, manager.cli_names)
            if manager.cli_path
            else (
                f"{', '.join(map(theme.invoked_command, manager.cli_names))} not found"
            ),
        )

        # Build up the version column content.
        version_infos = ""
        if manager.executable:
            version_infos = OK if manager.fresh else KO
            if manager.version:
                version_infos += f" {manager.version}"
                if not manager.fresh:
                    version_infos += f" {manager.requirement}"

        table.append(
            (
                getattr(theme, "success" if manager.fresh else "error")(manager.id),
                manager.name,
                os_infos,
                cli_infos,
                OK if manager.executable else "",
                version_infos,
            ),
        )

    ctx.find_root().print_table(
        (
            ("Manager ID", SortableField.MANAGER_ID),
            ("Name", SortableField.MANAGER_NAME),
            ("Supported", None),
            ("CLI", None),
            ("Executable", None),
            ("Version", SortableField.VERSION),
        ),
        table,
        ctx.obj.sort_by,
    )


@mpm.command(aliases=["list"], short_help="List installed packages.", section=EXPLORE)
@option(
    "-d",
    "--duplicates",
    is_flag=True,
    default=False,
    help="Only list installed packages sharing the same ID. Implies "
    "`--sort_by package_id` to make duplicates easier to compare between themselves.",
)
@pass_context
def installed(ctx, duplicates):
    """List all packages installed on the system by each manager."""
    # Build-up a global dict of installed packages per manager.
    installed_data = {}
    fields = (
        "id",
        "name",
        "installed_version",
    )

    for manager in ctx.obj.selected_managers(implements_operation=Operations.installed):
        packages = tuple(packages_asdict(manager.installed, fields))

        installed_data[manager.id] = {
            "id": manager.id,
            "name": manager.name,
            "packages": packages,
        }

        # Serialize errors at the last minute to gather all we encountered.
        installed_data[manager.id]["errors"] = list(
            {expt.error for expt in manager.cli_errors},
        )

    # Filters out non-duplicate packages.
    if duplicates:
        # Re-group packages by their IDs.
        package_sources = {}
        for manager_id, installed_pkg in installed_data.items():
            for package in installed_pkg["packages"]:
                package_sources.setdefault(package["id"], set()).add(manager_id)
        logging.debug(f"Managers sourcing each package: {package_sources}")

        # Identify package IDs shared by multiple managers.
        duplicates_ids = {
            pid for pid, managers in package_sources.items() if len(managers) > 1
        }
        logging.debug(f"Duplicates: {duplicates_ids}")

        # Remove non-duplicates from results.
        for manager_id, manager_data in installed_data.items():
            duplicate_packages = tuple(
                p for p in manager_data["packages"] if p["id"] in duplicates_ids
            )
            manager_data["packages"] = duplicate_packages

    # Machine-friendly data rendering.
    if ctx.meta["click_extra.table_format"] == TableFormat.JSON:
        print_json(installed_data)
        ctx.exit()

    # Human-friendly content rendering.
    table = []
    for manager_id, installed_pkg in installed_data.items():
        table += [
            (
                info["id"],
                info["name"],
                manager_id,
                info["installed_version"] if info["installed_version"] else "?",
            )
            for info in installed_pkg["packages"]
        ]

    # Force sorting by package ID in duplicate mode.
    sort_by = ctx.obj.sort_by
    if duplicates:
        logging.info(
            "Force table sorting on package ID because of --duplicates option."
        )
        sort_by = SortableField.PACKAGE_ID

    # Print table.
    ctx.find_root().print_table(
        (
            ("Package ID", SortableField.PACKAGE_ID),
            ("Name", SortableField.PACKAGE_NAME),
            ("Manager", SortableField.MANAGER_ID),
            ("Installed version", SortableField.VERSION),
        ),
        table,
        sort_by,
    )

    if ctx.obj.stats:
        print_stats(Counter({k: len(v["packages"]) for k, v in installed_data.items()}))


@mpm.command(short_help="List outdated packages.", section=EXPLORE)
@option(
    "--plugin-output",
    is_flag=True,
    default=False,
    help="Output results for direct consumption by an Xbar/SwiftBar-compatible plugin. "
    "The layout is dynamic and depends on environment variables set by either Xbar "
    "or SwiftBar.",
)
@pass_context
def outdated(ctx, plugin_output):
    """List available package upgrades and their versions for each manager."""
    # Build-up a global list of outdated packages per manager.
    outdated_data = {}
    fields = (
        "id",
        "name",
        "installed_version",
        "latest_version",
    )

    for manager in ctx.obj.selected_managers(implements_operation=Operations.outdated):
        packages = tuple(packages_asdict(manager.outdated, fields))

        outdated_data[manager.id] = {
            "id": manager.id,
            "name": manager.name,
            "packages": packages,
        }

        # Serialize errors at the last minute to gather all we encountered.
        outdated_data[manager.id]["errors"] = list(
            {expt.error for expt in manager.cli_errors},
        )

    # Machine-friendly data rendering.
    if ctx.meta["click_extra.table_format"] == TableFormat.JSON:
        print_json(outdated_data)
        ctx.exit()

    # Xbar/SwiftBar-friendly plugin rendering.
    if plugin_output:
        BarPluginRenderer().print(outdated_data)
        ctx.exit()

    # Human-friendly content rendering.
    table = []
    for manager_id, outdated_pkg in outdated_data.items():
        for info in outdated_pkg["packages"]:
            installed_version, latest_version = colored_diff(
                info["installed_version"] if info["installed_version"] else "?",
                info["latest_version"],
            )
            table.append(
                (
                    info["id"],
                    info["name"],
                    manager_id,
                    installed_version,
                    latest_version,
                ),
            )

    # Sort and print table.
    ctx.find_root().print_table(
        (
            ("Package ID", SortableField.PACKAGE_ID),
            ("Name", SortableField.PACKAGE_NAME),
            ("Manager", SortableField.MANAGER_ID),
            ("Installed version", SortableField.VERSION),
            ("Latest version", None),
        ),
        table,
        ctx.obj.sort_by,
    )

    if ctx.obj.stats:
        print_stats(Counter({k: len(v["packages"]) for k, v in outdated_data.items()}))


# TODO: make it a --search-strategy=[exact, fuzzy, extended]
# Add details helps => exact: is case-sensitive, and keep all non-alnum chars
# fuzzy: query is case-insensitive, stripped-out of non-alnum chars and
# tokenized (no order sensitive)
# extended, same as fuzzy, but do not limit search to package ID and name.
# extended to description and other metadata depending on manager support.
# Modes:
#  1. strict (--exact, on ID or name)
#  2. substring (regex, no case, no split)
#  3. fuzzy (token-based)
#  4. extended (fuzzy + metadata)
@mpm.command(short_help="Search packages.", section=EXPLORE)
@option(
    "--extended/--id-name-only",
    default=False,
    help="Extend search to description, instead of restricting it to package ID and "
    "name. Implies --description.",
)
@option(
    "--exact/--fuzzy",
    default=False,
    help="Only returns exact matches on package ID or name.",
)
@option(
    "--refilter/--no-refilter",
    default=True,
    help="Let mpm refilters managers' search results.",
)
@argument("query", type=STRING, required=True)
@pass_context
def search(ctx, extended, exact, refilter, query):
    """Search each manager for a package ID, name or description matching the query."""
    # --extended implies --description.
    show_description = ctx.obj.description
    if extended and not ctx.obj.description:
        logging.info("--extended option forces --description option.")
        show_description = True

    # Build-up a global list of package matches per manager.
    matches = {}
    fields = (
        "id",
        "name",
        "latest_version",
        "description",
    )

    search_method = "refiltered_search" if refilter else "search"
    for manager in ctx.obj.selected_managers(implements_operation=Operations.search):
        packages = tuple(
            packages_asdict(
                getattr(manager, search_method)(query, extended, exact),
                fields,
            ),
        )

        matches[manager.id] = {
            "id": manager.id,
            "name": manager.name,
            "packages": packages,
        }

        # Serialize errors at the last minute to gather all we encountered.
        matches[manager.id]["errors"] = list(
            {expt.error for expt in manager.cli_errors},
        )

    # Machine-friendly data rendering.
    if ctx.meta["click_extra.table_format"] == TableFormat.JSON:
        print_json(matches)
        ctx.exit()

    # Prepare highlighting helpers.
    query_parts = {query}.union(PackageManager.query_parts(query))
    highlight_query = cached(LRI(max_size=1000))(
        partial(
            highlight,
            patterns=query_parts,
            styling_func=theme.search,
            ignore_case=True,
        ),
    )

    # Human-friendly content rendering.
    table = []
    for manager_id, matching_pkg in matches.items():
        for pkg in matching_pkg["packages"]:
            line = [
                highlight_query(pkg["id"]) if pkg["id"] else "",
                highlight_query(pkg["name"]) if pkg["name"] else "",
                manager_id,
                pkg["latest_version"] if pkg["latest_version"] else "?",
            ]
            if show_description:
                line.append(
                    highlight_query(pkg.get("description"))
                    if pkg.get("description")
                    else "",
                )
            table.append(line)

    # Sort and print table.
    headers = [
        ("Package ID", SortableField.PACKAGE_ID),
        ("Name", SortableField.PACKAGE_NAME),
        ("Manager", SortableField.MANAGER_ID),
        ("Latest version", SortableField.VERSION),
    ]
    if show_description:
        headers.append(("Description", None))
    ctx.find_root().print_table(headers, table, ctx.obj.sort_by)

    if ctx.obj.stats:
        print_stats(Counter({k: len(v["packages"]) for k, v in matches.items()}))


@mpm.command(aliases=["locate"], short_help="Locate CLIs on system.", section=EXPLORE)
@argument("cli_names", type=STRING, nargs=-1, required=True)
@pass_context
def which(ctx, cli_names):
    """Search from the user's environment all CLIs matching the query.

    This is mpm's own version of the `which -a` UNIX command, used internally to locate
    binaries for each manager. It is exposed as a subcommand for convenience and to help
    troubleshoot CLI resolution logic.

    Compared to the venerable `which` command, this will respect the additional path
    configured for each package manager. It will ignore files that are empty (0 size).
    On Windows, it additionally suppress the default lookup in the current directory,
    which takes precedence on other paths.
    """
    if ctx.obj.sort_by:
        logging.warning("Ignore --sort-by option for which command.")

    # Machine-friendly data rendering.
    if ctx.meta["click_extra.table_format"] == TableFormat.JSON:
        cli_data = [
            {
                "manager_id": manager.id,
                "cli_paths": list(manager.search_all_cli(cli_names)),
            }
            for manager in ctx.obj.selected_managers()
        ]
        print_json(cli_data)
        ctx.exit()

    # Print table.
    table = []
    for manager in ctx.obj.selected_managers():
        for priority, found_cli in enumerate(manager.search_all_cli(cli_names)):
            # Resolve symlinks and highlight the CLI name.
            symlink = ""
            if found_cli.is_symlink():
                symlink = "→ " + highlight_cli_name(found_cli.resolve(), cli_names)
            table.append(
                (
                    manager.id,
                    str(priority),
                    highlight_cli_name(found_cli, cli_names),
                    symlink,
                ),
            )
    ctx.find_root().print_table(
        (
            ("Manager ID", SortableField.MANAGER_ID),
            ("Priority", None),
            ("CLI path", None),
            ("Symlink destination", None),
        ),
        table,
        ctx.obj.sort_by,
    )


@mpm.command(short_help="Install a package.", section=MAINTENANCE)
@argument(
    "packages_specs",
    type=STRING,
    nargs=-1,
    required=True,
    help="A mix of plain <package_id>, simple <package_id@version> specifiers or full "
    "<pkg:npm/left-pad> purls.",
)
# TODO: add a --force/--reinstall flag
@pass_context
def install(ctx, packages_specs):
    """Install one or more packages.

    This subcommand is sensible to the order of the package managers selected by the
    user.

    Installation will first proceed for all the packages found to be tied to a specific
    manager. Which is the case for packages provided with precise package specifiers
    (like purl). This will also happens in situations in which a tighter selection of
    managers is provided by the user.

    For packages whose manager is not known, or if multiple managers are candidates for
    the installation, mpm will try to find the best manager to install it with.

    Installation will be attempted with each manager, in the order they were selected.
    If a search for the package ID returns no result from the highest-priority manager,
    we will skip the installation and try the next available managers in the order of
    their priority.
    """
    # Cast generator to tuple because of reuse.
    selected_managers = tuple(
        ctx.obj.selected_managers(implements_operation=Operations.install),
    )
    manager_ids = tuple(manager.id for manager in selected_managers)
    logging.info(
        "Installation priority: > "
        f"{' > '.join(map(theme.invoked_command, manager_ids))}",
    )

    solver = Solver(packages_specs, manager_priority=manager_ids)
    packages_per_managers = solver.resolve_specs_group_by_managers()

    # Install all packages deterministiccaly tied to a specific manager.
    for manager_id, package_specs in packages_per_managers.items():
        if not manager_id:
            continue
        for spec in package_specs:
            try:
                logging.info(
                    f"Install {spec} package with "
                    f"{theme.invoked_command(manager_id)}...",
                )
                manager = pool.get(manager_id)
                output = manager.install(spec.package_id, version=spec.version)
            except NotImplementedError:
                logging.warning(
                    f"{theme.invoked_command(manager_id)} "
                    "does not implement install operation.",
                )
                continue
            echo(output)

    unmatched_packages = packages_per_managers.get(None, [])
    for spec in unmatched_packages:
        for manager in selected_managers:
            logging.info(
                f"Try to install {spec} with {theme.invoked_command(manager.id)}.",
            )

            # Is the package available on this manager?
            matches = None
            try:
                matches = tuple(
                    manager.refiltered_search(
                        extended=False,
                        exact=True,
                        query=spec.package_id,
                    ),
                )
            except NotImplementedError:
                logging.warning(
                    f"{theme.invoked_command(manager.id)} "
                    "does not implement search operation.",
                )
                logging.info(
                    f"{spec.package_id} existence unconfirmed, "
                    "try to directly install it...",
                )
            else:
                if not matches:
                    logging.warning(
                        f"No {spec.package_id} package found "
                        f"on {theme.invoked_command(manager.id)}.",
                    )
                    continue
                # Prevents any incomplete or bad implementation of exact search.
                if len(matches) != 1:
                    msg = "Exact search returned multiple packages."
                    raise ValueError(msg)

            # Allow install subcommand to fail to have the opportunity to catch the
            # CLIError exception and print a comprehensive message.
            with patch.object(manager, "stop_on_error", True):
                try:
                    logging.info(
                        f"Install {spec} package "
                        f"with {theme.invoked_command(manager.id)}...",
                    )
                    output = manager.install(spec.package_id, version=spec.version)
                except NotImplementedError:
                    logging.warning(
                        f"{theme.invoked_command(manager.id)} "
                        "does not implement install operation.",
                    )
                    continue
                except CLIError:
                    logging.warning(
                        f"Could not install {spec} "
                        f"with {theme.invoked_command(manager.id)}.",
                    )
                    continue

            echo(output)
            ctx.exit()


@mpm.command(aliases=["update"], short_help="Upgrade packages.", section=MAINTENANCE)
@option(
    "-A",
    "--all",
    is_flag=True,
    default=False,
    help="Upgrade all outdated packages. "
    "Will make the command ignore package IDs provided as parameters.",
)
@argument(
    "packages_specs",
    type=STRING,
    nargs=-1,
    help="A mix of plain <package_id>, simple <package_id@version> specifiers or full "
    "<pkg:npm/left-pad> purls.",
)
@pass_context
def upgrade(ctx, all, packages_specs):
    """Upgrade one or more outdated packages.

    All outdated package will be upgraded by default if no specifiers are provided as
    arguments. I.e. assumes -A/--all option if no [PACKAGES_SPECS]....

    Packages recognized by multiple managers will be upgraded with each of them. You can
    fine-tune this behavior with more precise package specifiers (like purl) and/or
    tighter selection of managers.

    Packages unrecognized by any selected manager will be skipped.
    """
    if not all and not packages_specs:
        logging.info("No package provided, assume -A/--all option.")
        all = True

    # Full upgrade.
    if all:
        if packages_specs:
            # Deduplicate and sort specifiers for terseness.
            logging.warning(
                f"Ignore {', '.join(sorted(set(packages_specs)))} specifiers "
                "and proceed to a full upgrade...",
            )
        for manager in ctx.obj.selected_managers(
            implements_operation=Operations.upgrade_all,
        ):
            logging.info(
                "Upgrade all outdated packages "
                f"from {theme.invoked_command(manager.id)}...",
            )
            output = manager.upgrade()
            if output:
                logging.info(output)
        ctx.exit()

    # Cast generator to tuple because of reuse.
    selected_managers = tuple(
        ctx.obj.selected_managers(implements_operation=Operations.upgrade),
    )
    manager_ids = tuple(manager.id for manager in selected_managers)

    # Get the subset of selected managers that are implementing the installed operation,
    # so we can query it and know if a package has been installed with it.
    sourcing_managers = tuple(
        ctx.obj.selected_managers(
            keep=manager_ids,
            implements_operation=Operations.installed,
        ),
    )

    solver = Solver(packages_specs, manager_priority=manager_ids)
    for package_id, spec in solver.resolve_package_specs():
        source_manager_ids = set()
        # Use the manager from the spec.
        if spec.manager_id:
            source_manager_ids.add(spec.manager_id)
        # Package is not bound to a manager by the user's specifiers.
        else:
            logging.info(
                f"{spec} not tied to a manager. Search all managers recognizing it.",
            )
            # Find all the managers that have the package installed.
            for manager in sourcing_managers:
                if package_id in map(attrgetter("id"), manager.installed):
                    logging.info(
                        f"{package_id} has been installed "
                        f"with {theme.invoked_command(manager.id)}.",
                    )
                    source_manager_ids.add(manager.id)

        if not source_manager_ids:
            logging.error(
                f"{package_id} is not recognized by any of the selected manager. "
                "Skip it.",
            )
            continue

        logging.info(
            f"Upgrade {package_id} "
            f"with {', '.join(map(theme.invoked_command, sorted(source_manager_ids)))}",
        )
        for manager_id in source_manager_ids:
            manager = pool.get(manager_id)
            output = manager.upgrade(package_id, version=spec.version)
            if output:
                logging.info(output)


@mpm.command(aliases=["uninstall"], short_help="Remove a package.", section=MAINTENANCE)
@argument(
    "packages_specs",
    type=STRING,
    nargs=-1,
    required=True,
    help="A mix of plain <package_id>, simple <package_id@version> specifiers or full "
    "<pkg:npm/left-pad> purls.",
)
@pass_context
def remove(ctx, packages_specs):
    """Remove one or more packages.

    Packages recognized by multiple managers will be remove with each of them. You can
    fine-tune this behavior with more precise package specifiers (like purl) and/or
    tighter selection of managers.

    Packages unrecognized by any selected manager will be skipped.
    """
    # Cast generator to tuple because of reuse.
    selected_managers = tuple(
        ctx.obj.selected_managers(implements_operation=Operations.remove),
    )
    manager_ids = tuple(manager.id for manager in selected_managers)

    # Get the subset of selected managers that are implementing the installed operation,
    # so we can query it and know if a package has been installed with it.
    sourcing_managers = tuple(
        ctx.obj.selected_managers(
            keep=manager_ids,
            implements_operation=Operations.installed,
        ),
    )

    solver = Solver(packages_specs, manager_priority=manager_ids)
    for package_id, spec in solver.resolve_package_specs():
        source_manager_ids = set()
        # Use the manager from the spec.
        if spec.manager_id:
            source_manager_ids.add(spec.manager_id)
        # Package is not bound to a manager by the user's specifiers.
        else:
            logging.info(
                f"{spec} not tied to a manager. Search all managers recognizing it.",
            )
            # Find all the managers that have the package installed.
            for manager in sourcing_managers:
                if package_id in map(attrgetter("id"), manager.installed):
                    logging.info(
                        f"{package_id} has been installed "
                        f"with {theme.invoked_command(manager.id)}.",
                    )
                    source_manager_ids.add(manager.id)

        if not source_manager_ids:
            logging.error(
                f"{package_id} is not recognized by any of the selected manager. "
                "Skip it.",
            )
            continue

        logging.info(
            f"Remove {package_id} "
            f"with {', '.join(map(theme.invoked_command, sorted(source_manager_ids)))}",
        )
        for manager_id in source_manager_ids:
            manager = pool.get(manager_id)
            output = manager.remove(package_id)
            if output:
                logging.info(output)


@mpm.command(short_help="Sync local package info.", section=MAINTENANCE)
@pass_context
def sync(ctx):
    """Sync local package metadata and info from external sources."""
    for manager in ctx.obj.selected_managers(implements_operation=Operations.sync):
        logging.info(f"Sync {theme.invoked_command(manager.id)} package info...")
        manager.sync()


@mpm.command(short_help="Cleanup local data.", section=MAINTENANCE)
@pass_context
def cleanup(ctx):
    """Cleanup local data, temporary artifacts and removes orphaned dependencies."""
    for manager in ctx.obj.selected_managers(implements_operation=Operations.cleanup):
        logging.info(f"Cleanup {theme.invoked_command(manager.id)}...")
        manager.cleanup()


@mpm.command(
    aliases=["lock", "freeze", "snapshot"],
    short_help="Save installed packages to a TOML file.",
    section=SNAPSHOTS,
)
@option(
    "--overwrite",
    "--force",
    "--replace",
    is_flag=True,
    default=False,
    help="Allow the provided TOML file to be silently wiped out if it already exists.",
)
@option(
    "--merge",
    is_flag=True,
    default=False,
    help="Read the provided TOML file and update each entry with the version currently "
    "installed on the system. Requires the [TOML_PATH] argument.",
)
@option(
    "--update-version",
    is_flag=True,
    default=False,
    help="Read the provided TOML file and update each existing entry with the version "
    "currently installed on the system. Requires the [TOML_PATH] argument.",
)
@argument(
    "toml_path",
    type=file_path(writable=True, resolve_path=True, allow_dash=True),
    default="-",
)
@pass_context
def backup(ctx, overwrite, merge, update_version, toml_path):
    """Dump the list of installed packages to a TOML file.

    By default the generated TOML content is displayed directly in the console output.
    So `mpm backup` is the same as a call to `mpm backup -`. To have the result written
    in a file on disk, specify the output file like so: `mpm backup packages.toml`.

    Files produced by this subcommand can be safely consumed by `mpm restore`.

    Sections of the TOML file will be named after the manager ID. Sections are ordered in
    the same order as the manager selection priority. Each section will contain a list of
    package IDs and their installed version.
    """
    if is_stdout(toml_path):
        if merge:
            logging.critical(
                "--merge requires the [TOML_PATH] argument to point to a file.",
            )
            ctx.exit(2)
        if update_version:
            logging.critical(
                "--update-version requires the [TOML_PATH] argument to point to a "
                "file.",
            )
            ctx.exit(2)
        if overwrite:
            logging.warning("Ignore the --overwrite/--force/--replace option.")
        logging.info(f"Print installed package list to {sys.stdout.name}")

    else:
        if merge and update_version:
            logging.critical("--merge and --update-version are mutually exclusive.")
            ctx.exit(2)

        if merge:
            logging.info(f"Merge all installed packages into {toml_path}")
        elif update_version:
            logging.info(
                f"Update in-place all versions of installed packages "
                f"found in {toml_path}",
            )
        else:
            logging.info(f"Dump all installed packages into {toml_path}")

        if toml_path.exists():
            if overwrite:
                logging.warning("Target file exist and will be overwritten.")
            else:
                if merge or update_version:
                    logging.warning("Ignore the --overwrite/--force/--replace option.")
                else:
                    logging.critical("Target file exist and will be overwritten.")
                    ctx.exit(2)
        elif merge:
            logging.critical("--merge requires an existing file.")
            ctx.exit(2)
        elif update_version:
            logging.critical("--update-version requires an existing file.")
            ctx.exit(2)

        if toml_path.suffix.lower() != ".toml":
            logging.critical("Target file is not a TOML file.")
            ctx.exit(2)

    installed_data = {}
    fields = (
        "id",
        "installed_version",
    )
    if merge or update_version:
        installed_data = tomllib.loads(toml_path.read_text(encoding="utf-8"))

    # Leave some metadata as comment.
    content = (
        f"# Generated by mpm v{__version__}.\n"
        f"# Timestamp: {datetime.now(tz=timezone.utc).isoformat()}.\n\n"
    )
    # Create one section for each manager.
    for manager in ctx.obj.selected_managers(implements_operation=Operations.installed):
        logging.info(f"Dumping packages from {theme.invoked_command(manager.id)}...")

        packages = tuple(packages_asdict(manager.installed, fields))

        for pkg in packages:
            # Only update version in that mode if the package is already referenced
            # into original TOML file.
            if update_version:
                if pkg["id"] in installed_data.get(manager.id, {}):
                    installed_data[manager.id][pkg["id"]] = str(
                        pkg["installed_version"],
                    )
            # Insert installed package in data structure for standard dump and merge
            # mode.
            else:
                installed_data.setdefault(manager.id, {})[pkg["id"]] = str(
                    pkg["installed_version"],
                )

        # Re-sort package list.
        if installed_data.get(manager.id):
            installed_data[manager.id] = dict(
                sorted(
                    installed_data[manager.id].items(),
                    # Case-insensitive lexicographical sort on keys.
                    key=lambda i: (i[0].lower(), i[0]),
                ),
            )

    # Write each section separated by an empty line for readability.
    content += "\n".join(
        (
            tomli_w.dumps({manager_id: packages})
            for manager_id, packages in installed_data.items()
        )
    )

    echo(content, file=prep_path(toml_path))

    if ctx.obj.stats:
        print_stats(Counter({k: len(v) for k, v in installed_data.items()}))


@mpm.command(
    short_help="Install packages referenced in TOML files.",
    section=SNAPSHOTS,
)
@argument("toml_files", type=File("r"), required=True, nargs=-1)
@pass_context
def restore(ctx, toml_files):
    """Read TOML files then install or upgrade each package referenced in them."""
    # TODO: add an artificial order for managers, so that the [cask] section can install
    # mas CLI first then have [mas] section work in one-go. Use-case: my dotfiles.
    for toml_input in toml_files:
        is_stdin = isinstance(toml_input, TextIOWrapper)
        if is_stdin:
            toml_input.reconfigure(encoding="utf-8")
            toml_filepath = toml_input.name
            toml_content = toml_input.read()
        else:
            toml_filepath = Path(toml_input.name).resolve()
            toml_content = toml_filepath.read_text(encoding="utf-8")

        logging.info(f"Load package list from {toml_filepath}")
        doc = tomllib.loads(toml_content)

        # List unrecognized sections.
        ignored_sections = [
            f"[{section}]" for section in doc if section not in pool.all_manager_ids
        ]
        if ignored_sections:
            plural = "s" if len(ignored_sections) > 1 else ""
            sections = ", ".join(ignored_sections)
            logging.info(f"Ignore {sections} section{plural}.")

        for manager in ctx.obj.selected_managers(
            implements_operation=Operations.install,
        ):
            if manager.id not in doc:
                logging.warning(
                    f"No [{theme.invoked_command(manager.id)}] section found.",
                )
                continue
            logging.info(f"Restore {theme.invoked_command(manager.id)} packages...")
            for package_id, version in doc[manager.id].items():
                spec = Specifier(
                    raw_spec=f"pkg:{manager.id}:/{package_id}{VERSION_SEP}{package_id}",
                    package_id=package_id,
                    manager_id=manager.id,
                    version=version,
                )
                logging.info(f"Install {spec}...")
                output = manager.install(spec.package_id, version=spec.version)
                echo(output)


@mpm.command(
    short_help="Export installed packages to a SBOM document.",
    section=SBOM_SECTION,
)
@option(
    "--spdx/--cyclonedx",
    default=True,
    help="SBOM standard to export to.",
)
@option(
    "--format",
    "export_format",
    type=EnumChoice(ExportFormat),
    help=f"File format of the export. Defaults to JSON for {sys.stdout.name}. If not "
    "provided, will be autodetected from file extension.",
)
@option(
    "--overwrite",
    "--force",
    "--replace",
    is_flag=True,
    default=False,
    help="Allow the target file to be silently wiped out if it already exists.",
)
@argument(
    "export_path",
    type=file_path(writable=True, resolve_path=True, allow_dash=True),
    default="-",
)
@pass_context
def sbom(ctx, spdx, export_format, overwrite, export_path):
    """Export list of installed packages to a SPDX or CycloneDX file."""
    standard = "SPDX" if spdx else "CycloneDX"

    if is_stdout(export_path):
        if overwrite:
            logging.warning("Ignore the --overwrite/--force/--replace option.")
        logging.info(f"Print {standard} export to {sys.stdout.name}")

    else:
        logging.info(f"Export installed packages in {standard} to {export_path}")
        if export_path.exists():
            msg = "Target file exist and will be overwritten."
            if overwrite:
                logging.warning(msg)
            else:
                logging.critical(msg)
                ctx.exit(2)

    # <stdout> format defaults to JSON.
    if is_stdout(export_path):
        if not export_format:
            export_format = ExportFormat.JSON
    # If no export format has been provided, guess it from file name.
    else:
        guessed_format = SBOM.autodetect_export_format(export_path)
        if not export_format:
            export_format = guessed_format
        else:
            if export_format != guessed_format:
                logging.critical(
                    f"Selected {export_format} does not match file extension."
                )
                ctx.exit(2)

    if spdx:
        sbom_class = SPDX
    else:
        if export_format not in (ExportFormat.JSON, ExportFormat.XML):
            logging.critical(f"{standard} does not support {export_format} format.")
            ctx.exit(2)
        sbom_class = CycloneDX

    sbom = sbom_class(export_format)
    sbom.init_doc()

    for manager in ctx.obj.selected_managers(implements_operation=Operations.installed):
        logging.info(f"Export packages from {theme.invoked_command(manager.id)}...")
        for package in manager.installed:
            sbom.add_package(manager, package)

    echo(sbom.export(), file=prep_path(export_path))
